import datetime
import random
import tempfile
from contextlib import contextmanager

import numpy
import pandas
import yaml
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine
import testing.postgresql
from triage.component.catwalk.db import ensure_db
from triage.component.catwalk.storage import MatrixStore, ProjectStorage
from triage.component.results_schema import Model, Matrix
from triage.experiments import CONFIG_VERSION
from tests.results_tests.factories import init_engine, session, MatrixFactory


@contextmanager
def fake_metta(matrix_dict, metadata):
    """Stores matrix and metadata in a metta-data-like form

    Args:
    matrix_dict (dict) of form { columns: values }.
        Expects an entity_id to be present which it will use as the index
    metadata (dict). Any metadata that should be set

    Yields:
        tuple of filenames for matrix and metadata
    """
    matrix = pandas.DataFrame.from_dict(matrix_dict).set_index('entity_id')
    with tempfile.NamedTemporaryFile() as matrix_file:
        with tempfile.NamedTemporaryFile('w') as metadata_file:
            hdf = pandas.HDFStore(matrix_file.name)
            hdf.put('title', matrix, data_columns=True)
            matrix_file.seek(0)

            yaml.dump(metadata, metadata_file)
            metadata_file.seek(0)
            yield (matrix_file.name, metadata_file.name)


def fake_labels(length):
    return numpy.array([random.choice([True, False]) for i in range(0, length)])


class MockTrainedModel(object):
    def predict_proba(self, dataset):
        return numpy.random.rand(len(dataset), len(dataset))

class MockMatrixStore(MatrixStore):
    def __init__(self, matrix_type, matrix_uuid, label_count, db_engine, init_labels=None, metadata_overrides=None, matrix=None):
        base_metadata = {
            'feature_start_time': datetime.date(2014, 1, 1),
            'end_time': datetime.date(2015, 1, 1),
            'as_of_date_frequency': '1y',
            'matrix_id': 'some_matrix',
            'label_name': 'label',
            'label_timespan': '3month',
            'indices': ['entity_id'],
            'matrix_type': matrix_type
        }
        metadata_overrides = metadata_overrides or {}
        base_metadata.update(metadata_overrides)
        if matrix is None:
            matrix = pandas.DataFrame.from_dict({
                'entity_id': [1, 2],
                'feature_one': [3, 4],
                'feature_two': [5, 6],
                'label': [7, 8]
            }).set_index('entity_id')
        if init_labels is None:
            init_labels = []
        self.matrix = matrix
        self.metadata = base_metadata
        self.label_count = label_count
        self.init_labels = init_labels
        self.matrix_uuid = matrix_uuid

        session = sessionmaker(db_engine)()
        session.add(Matrix(matrix_uuid=matrix_uuid))

    def labels(self):
        if self.init_labels == []:
            return fake_labels(self.label_count)
        else:
            return self.init_labels


def fake_trained_model(db_engine, train_matrix_uuid='efgh'):
    """Creates and stores a trivial trained model and training matrix

    Args:
        db_engine (sqlalchemy.engine)

    Returns:
        (int) model id for database retrieval
    """
    session = sessionmaker(db_engine)()
    session.merge(Matrix(matrix_uuid=train_matrix_uuid))

    # Create the fake trained model and store in db
    trained_model = MockTrainedModel()
    db_model = Model(model_hash='abcd', train_matrix_uuid=train_matrix_uuid)
    session.add(db_model)
    session.commit()
    return trained_model, db_model.model_id


def matrix_metadata_creator(**override_kwargs):
    """Create a sample valid matrix metadata with optional overrides

    Args:
    **override_kwargs: Keys and values to override in the metadata

    Returns: (dict)
    """
    base_metadata = {
        'feature_start_time': datetime.date(2012, 12, 20),
        'end_time': datetime.date(2016, 12, 20),
        'label_name': 'label',
        'as_of_date_frequency': '1w',
        'max_training_history': '5y',
        'matrix_id': 'tester-1',
        'state': 'active',
        'cohort_name': 'default',
        'label_timespan': '1y',
        'metta-uuid': '1234',
        'matrix_type': 'test',
        'feature_names': ['ft1', 'ft2'],
        'feature_groups': ['all: True'],
        'indices': ['entity_id', 'as_of_date'],
    }
    for override_key, override_value in override_kwargs.items():
        base_metadata[override_key] = override_value

    return base_metadata


def matrix_creator(index=None):
    """Return a sample matrix.

    Args:
        index (list, optional): The matrix index column names. Defaults to ['entity_id', 'date']
    """
    if not index:
        index = ['entity_id', 'as_of_date']
    source_dict = {
        'entity_id': [1, 2],
        'feature_one': [3, 4],
        'feature_two': [5, 6],
        'label': [0, 1]
    }
    if 'as_of_date' in index:
        source_dict['as_of_date'] = [datetime.datetime(2016, 1, 1), datetime.datetime(2016, 1, 1)]
    return pandas.DataFrame.from_dict(source_dict).set_index(index)


def get_matrix_store(project_storage, matrix=None, metadata=None):
    """Return a matrix store associated with the given project storage. Also adds an entry in the matrices table if it doesn't exist already

    Args:
        project_storage (triage.component.catwalk.storage.ProjectStorage) A project's storage
        matrix (dataframe, optional): A matrix to store. Defaults to the output of matrix_creator()
        metadata (dict, optional): matrix metadata. defaults to the output of matrix_metadata_creator()
    """
    if matrix is None:
        matrix = matrix_creator()
    if not metadata:
        metadata = matrix_metadata_creator()
    matrix_store = project_storage.matrix_storage_engine().get_store(metadata['metta-uuid'])
    matrix_store.matrix = matrix
    matrix_store.metadata = metadata
    matrix_store.save()
    if session.query(Matrix).filter(Matrix.matrix_uuid == matrix_store.uuid).count() == 0:
        MatrixFactory(matrix_uuid=matrix_store.uuid)
        session.commit()
    return matrix_store


@contextmanager
def rig_engines():
    """Set up a db engine and project storage engine

    Yields (tuple) (database engine, project storage engine)
    """
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        ensure_db(db_engine)
        init_engine(db_engine)
        project_path = 'econ-dev/inspections'
        with tempfile.TemporaryDirectory() as temp_dir:
            project_storage = ProjectStorage(temp_dir)
            yield db_engine, project_storage


def populate_source_data(db_engine):
    complaints = [
        (1, '2010-10-01', 5),
        (1, '2011-10-01', 4),
        (1, '2011-11-01', 4),
        (1, '2011-12-01', 4),
        (1, '2012-02-01', 5),
        (1, '2012-10-01', 4),
        (1, '2013-10-01', 5),
        (2, '2010-10-01', 5),
        (2, '2011-10-01', 5),
        (2, '2011-11-01', 4),
        (2, '2011-12-01', 4),
        (2, '2012-02-01', 6),
        (2, '2012-10-01', 5),
        (2, '2013-10-01', 6),
        (3, '2010-10-01', 5),
        (3, '2011-10-01', 3),
        (3, '2011-11-01', 4),
        (3, '2011-12-01', 4),
        (3, '2012-02-01', 4),
        (3, '2012-10-01', 3),
        (3, '2013-10-01', 4),
    ]

    entity_zip_codes = [
        (1, '60120'),
        (2, '60123'),
        (3, '60123'),
    ]

    zip_code_events = [
        ('60120', '2012-10-01', 1),
        ('60123', '2012-10-01', 10),
    ]

    events = [
        (1, 1, '2011-01-01'),
        (1, 1, '2011-06-01'),
        (1, 1, '2011-09-01'),
        (1, 1, '2012-01-01'),
        (1, 1, '2012-01-10'),
        (1, 1, '2012-06-01'),
        (1, 1, '2013-01-01'),
        (1, 0, '2014-01-01'),
        (1, 1, '2015-01-01'),
        (2, 1, '2011-01-01'),
        (2, 1, '2011-06-01'),
        (2, 1, '2011-09-01'),
        (2, 1, '2012-01-01'),
        (2, 1, '2013-01-01'),
        (2, 1, '2014-01-01'),
        (2, 1, '2015-01-01'),
        (3, 0, '2011-01-01'),
        (3, 0, '2011-06-01'),
        (3, 0, '2011-09-01'),
        (3, 0, '2012-01-01'),
        (3, 0, '2013-01-01'),
        (3, 1, '2014-01-01'),
        (3, 0, '2015-01-01'),
    ]

    states = [
        (1, 'state_one', '2012-01-01', '2016-01-01'),
        (1, 'state_two', '2013-01-01', '2016-01-01'),
        (2, 'state_one', '2012-01-01', '2016-01-01'),
        (2, 'state_two', '2013-01-01', '2016-01-01'),
        (3, 'state_one', '2012-01-01', '2016-01-01'),
        (3, 'state_two', '2013-01-01', '2016-01-01'),
    ]

    db_engine.execute('''create table cat_complaints (
        entity_id int,
        as_of_date date,
        cat_sightings int
        )''')

    db_engine.execute('''create table entity_zip_codes (
        entity_id int,
        zip_code text
        )''')

    for entity_zip_code in entity_zip_codes:
        db_engine.execute(
            "insert into entity_zip_codes values (%s, %s)", entity_zip_code
        )

    db_engine.execute('''create table zip_code_events (
        zip_code text,
        as_of_date date,
        num_events int
    )''')
    for zip_code_event in zip_code_events:
        db_engine.execute(
            "insert into zip_code_events values (%s, %s, %s)", zip_code_event
        )

    for complaint in complaints:
        db_engine.execute(
            "insert into cat_complaints values (%s, %s, %s)",
            complaint
        )

    db_engine.execute('''create table events (
        entity_id int,
        outcome int,
        outcome_date date
    )''')

    for event in events:
        db_engine.execute(
            "insert into events values (%s, %s, %s)",
            event
        )

    db_engine.execute('''create table states (
        entity_id int,
        state text,
        start_time timestamp,
        end_time timestamp
    )''')

    for state in states:
        db_engine.execute(
            'insert into states values (%s, %s, %s, %s)',
            state
        )


def sample_config():
    temporal_config = {
        'feature_start_time': '2010-01-01',
        'feature_end_time': '2014-01-01',
        'label_start_time': '2011-01-01',
        'label_end_time': '2014-01-01',
        'model_update_frequency': '1year',
        'training_label_timespans': ['6months'],
        'test_label_timespans': ['6months'],
        'training_as_of_date_frequencies': '1day',
        'test_as_of_date_frequencies': '3months',
        'max_training_histories': ['6months'],
        'test_durations': ['1months'],
    }

    scoring_config = {
        'testing_metric_groups': [
            {'metrics': ['precision@'], 'thresholds': {'top_n': [2]}}
        ],
        'training_metric_groups': [
            {'metrics': ['precision@'], 'thresholds': {'top_n': [3]}}
        ]

    }

    grid_config = {
        'sklearn.tree.DecisionTreeClassifier': {
            'min_samples_split': [10, 100],
            'max_depth': [3,5],
            'criterion': ['gini']
        }
    }

    feature_config = [{
        'prefix': 'entity_features',
        'from_obj': 'cat_complaints',
        'knowledge_date_column': 'as_of_date',
        'aggregates_imputation': {'all': {'type': 'constant', 'value': 0}},
        'aggregates': [{
            'quantity': 'cat_sightings',
            'metrics': ['count', 'avg'],
        }],
        'intervals': ['1year'],
        'groups': ['entity_id']
    }, {
        'prefix': 'zip_code_features',
        'from_obj': 'entity_zip_codes join zip_code_events using (zip_code)',
        'knowledge_date_column': 'as_of_date',
        'aggregates_imputation': {'all': {'type': 'constant', 'value': 0}},
        'aggregates': [{
            'quantity': 'num_events',
            'metrics': ['max', 'min'],
        }],
        'intervals': ['1year'],
        'groups': ['entity_id', 'zip_code']
    }]

    cohort_config = {
        'dense_states': {
            'table_name': 'states',
            'state_filters': ['state_one or state_two'],
        }
    }

    label_config = {
        'query': """
            select
            events.entity_id,
            bool_or(outcome::bool)::integer as outcome
            from events
            where '{as_of_date}'::date <= outcome_date
                and outcome_date < '{as_of_date}'::date + interval '{label_timespan}'
                group by entity_id
        """,
        'name': 'custom_label_name',
        'include_missing_labels_in_train_as': False,
    }

    return {
        'config_version': CONFIG_VERSION,
        'label_config': label_config,
        'entity_column_name': 'entity_id',
        'model_comment': 'test2-final-final',
        'model_group_keys': ['label_name', 'label_type', 'custom_key'],
        'feature_aggregations': feature_config,
        'cohort_config': cohort_config,
        'temporal_config': temporal_config,
        'grid_config': grid_config,
        'scoring': scoring_config,
        'user_metadata': {'custom_key': 'custom_value'},
        'individual_importance': {'n_ranks': 2}
    }



import logging
import os
import pandas as pd
import numpy as np

from .plotting import plot_cats
from .utils import str_in_sql


class ModelGroupPerformancePlotter(object):

    def __init__(self, distance_from_best_table, directory=None):
        """Generate a plot illustrating the performance of a model group over time

        Args:
            distance_from_best_table (audition.DistanceFromBestTable)
                A pre-populated distance-from-best database table
        """
        self.distance_from_best_table = distance_from_best_table
        self.directory = directory

    def plot_all(self, metric_filters, model_group_ids, train_end_times):
        """For each metric, plot the value of that metric over time

        Arguments:
            metric_filters (list) The metrics to plot. Each element should be
                a dict with the following keys:

                metric (string) -- model evaluation metric, such as 'precision@'
                parameter (string) -- model evaluation metric parameter,
                    such as '300_abs'
            model_group_ids (list) - Model group ids to include in the plot
            train_end_times (list) - Train end times to include in the plot

        """
        for metric_filter in metric_filters:
            logging.info('Plotting model group performance for %s, %s',
                         metric_filter, train_end_times)
            df = self.generate_plot_data(
                metric=metric_filter['metric'],
                parameter=metric_filter['parameter'],
                model_group_ids=model_group_ids,
                train_end_times=train_end_times
            )
            self.plot(
                metric=metric_filter['metric'],
                parameter=metric_filter['parameter'],
                df_metric=df,
                train_end_times=train_end_times,
                directory=self.directory
            )

    def generate_plot_data(self, metric, parameter, model_group_ids, train_end_times):
        """Fetch data necessary for producing the plot from the distance table

        Arguments:
            metric (string) -- model evaluation metric, such as 'precision@'
            parameter (string) -- model evaluation metric parameter,
                such as '300_abs'
            model_group_ids (list) - Model group ids to include in the dataset
            train_end_times (list) - Train end times to include in the dataset

        Returns: (pandas.DataFrame) The relevant models and their performance
        on the given metric over time
        """

        base_df = pd.read_sql(
            '''
            select
                model_group_id,
                metric,
                parameter,
                train_end_time,
                raw_value,
                mg.model_type
            from {dist_table} dist
            join model_metadata.model_groups mg using (model_group_id)
            where model_group_id in ({model_group_ids})
            union
            select
                0 model_group_id,
                metric,
                parameter,
                train_end_time,
                best_case,
                'best case' model_type
            from {dist_table}
            group by 1, 2, 3, 4, 5, 6
            '''.format(
                dist_table=self.distance_from_best_table.distance_table,
                model_group_ids=str_in_sql(model_group_ids)
            ),
            self.distance_from_best_table.db_engine
        )
        df = base_df[
            (base_df['train_end_time'].isin(train_end_times)) &
            (base_df['metric'] == metric) &
            (base_df['parameter'] == parameter)
        ]
        return df

    def plot(self, metric, parameter, df_metric, train_end_times, directory, **plt_format_args):
        """Draw the plot representing the given data

        Arguments:
            metric (string) -- model evaluation metric, such as 'precision@'
            parameter (string) -- model evaluation metric parameter, such as '300_abs'
            df_metric (pandas.DataFrame)
            train_end_times (list) - Train end times to use for ticks
            **plt_format_args -- formatting arguments passed through to plot_cats()
        """
        cat_col = 'model_type'
        plt_title = '{} {} over time'.format(metric, parameter)

        # when setting the ticks, matplotlib sometimes has problems with datetimes given
        # as numpy.datetime64 objects, and converting from them to datetimes is ugly.
        # to get around this, we use the train_end_times given to the plot call as ticks
        # But to be defensive, we verify that these two versions of the list are the same
        for given_time, matrix_time in zip(
            train_end_times,
            sorted(df_metric['train_end_time'].unique()),
        ):
            given_time_as_numpy = np.datetime64(given_time)
            if given_time_as_numpy != matrix_time:
                raise ValueError(
                    'Train times given to the plotter do not match up with those '
                    'extracted from the database: '
                    '{} (given time) does not equal {} (matrix time)'
                    .format(given_time_as_numpy, matrix_time)
                )
        if directory:
            path_to_save = os.path.join(directory, f'metric_over_time_{metric}{parameter}.png')
        else:
            path_to_save = None

        plot_cats(
            frame=df_metric,
            x_col='train_end_time',
            y_col='raw_value',
            cat_col=cat_col,
            highlight_grp='best case',
            title=plt_title,
            x_label='train end time',
            y_label='value of {}'.format(metric),
            x_ticks=train_end_times,
            path_to_save=path_to_save,
            **plt_format_args
        )



"""Split results into model_metadata, test_results, and train_resultss

Revision ID: 89a8ce240bae
Revises: 7d57d1cf3429
Create Date: 2018-03-27 15:12:45.249609

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '89a8ce240bae'
down_revision = '7d57d1cf3429'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('CREATE SCHEMA IF NOT EXISTS model_metadata')
    op.execute('CREATE SCHEMA IF NOT EXISTS test_results')
    op.execute('CREATE SCHEMA IF NOT EXISTS train_results')
    op.execute('ALTER TABLE results.models SET SCHEMA model_metadata;' +
              ' ALTER TABLE results.model_groups SET SCHEMA model_metadata;' +
              ' ALTER TABLE results.experiments SET SCHEMA model_metadata;' +
              ' ALTER TABLE results.list_predictions SET SCHEMA model_metadata;' +
              ' ALTER TABLE results.predictions RENAME TO test_predictions;' +
              ' ALTER TABLE results.test_predictions SET SCHEMA test_results;' +
              ' ALTER TABLE results.evaluations RENAME TO test_evaluations;' +
              ' ALTER TABLE results.test_evaluations SET SCHEMA test_results;' +
              ' ALTER TABLE results.individual_importances SET SCHEMA test_results;' +
              ' ALTER TABLE results.feature_importances SET SCHEMA train_results;' )

    op.execute('ALTER TABLE model_metadata.models ADD COLUMN model_size real')
    op.create_table('matrices',
    sa.Column('matrix_id', sa.String(), nullable=True),
    sa.Column('matrix_uuid', sa.String(), nullable=False),
    sa.Column('matrix_type', sa.String(), nullable=True),
    sa.Column('labeling_window', sa.Interval(), nullable=True),
    sa.Column('n_examples', sa.Integer(), nullable=True),
    sa.Column('creation_time', sa.DateTime(timezone=True), server_default=sa.text('now()'), nullable=True),
    sa.Column('lookback_duration', sa.Interval(), nullable=True),
    sa.Column('feature_start_time', sa.DateTime(), nullable=True),
    sa.Column('matrix_metadata', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    sa.PrimaryKeyConstraint('matrix_uuid'),
    schema='model_metadata'
    )
    op.create_index(op.f('ix_model_metadata_matrices_matrix_uuid'), 'matrices', ['matrix_uuid'], unique=True, schema='model_metadata')

    #Create train_predictions and train_evaluations
    op.create_table('train_evaluations',
    sa.Column('model_id', sa.Integer(), nullable=False),
    sa.Column('evaluation_start_time', sa.DateTime(), nullable=False),
    sa.Column('evaluation_end_time', sa.DateTime(), nullable=False),
    sa.Column('as_of_date_frequency', sa.Interval(), nullable=False),
    sa.Column('metric', sa.String(), nullable=False),
    sa.Column('parameter', sa.String(), nullable=False),
    sa.Column('value', sa.Numeric(), nullable=True),
    sa.Column('num_labeled_examples', sa.Integer(), nullable=True),
    sa.Column('num_labeled_above_threshold', sa.Integer(), nullable=True),
    sa.Column('num_positive_labels', sa.Integer(), nullable=True),
    sa.Column('sort_seed', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['model_id'], ['model_metadata.models.model_id'], ),
    sa.PrimaryKeyConstraint('model_id', 'evaluation_start_time', 'evaluation_end_time', 'as_of_date_frequency', 'metric', 'parameter'),
    schema='train_results'
    )
    op.create_table('train_predictions',
    sa.Column('model_id', sa.Integer(), nullable=False),
    sa.Column('entity_id', sa.BigInteger(), nullable=False),
    sa.Column('as_of_date', sa.DateTime(), nullable=False),
    sa.Column('score', sa.Numeric(), nullable=True),
    sa.Column('label_value', sa.Integer(), nullable=True),
    sa.Column('rank_abs', sa.Integer(), nullable=True),
    sa.Column('rank_pct', sa.Float(), nullable=True),
    sa.Column('matrix_uuid', sa.Text(), nullable=True),
    sa.Column('test_label_timespan', sa.Interval(), nullable=True),
    sa.ForeignKeyConstraint(['matrix_uuid'], ['model_metadata.matrices.matrix_uuid'], ),
    sa.ForeignKeyConstraint(['model_id'], ['model_metadata.models.model_id'], ),
    sa.PrimaryKeyConstraint('model_id', 'entity_id', 'as_of_date'),
    schema='train_results'
    )

    # Foreign Keys for the models, test_predictions, and train_predictions tables
    op.execute('''ALTER TABLE model_metadata.models
                  ADD CONSTRAINT matrix_uuid_for_models
                  FOREIGN KEY (train_matrix_uuid)
                  REFERENCES model_metadata.matrices(matrix_uuid); ''')

    op.execute('''ALTER TABLE test_results.test_predictions
                  ADD CONSTRAINT matrix_uuid_for_testpred
                  FOREIGN KEY (matrix_uuid)
                  REFERENCES model_metadata.matrices(matrix_uuid); ''')

    op.execute('''ALTER TABLE train_results.train_predictions
                  ADD CONSTRAINT matrix_uuid_for_trainpred
                  FOREIGN KEY (matrix_uuid)
                  REFERENCES model_metadata.matrices(matrix_uuid); ''')

    op.execute("DROP SCHEMA IF EXISTS results")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('CREATE SCHEMA IF NOT EXISTS results')

    op.execute('''ALTER TABLE model_metadata.models DROP CONSTRAINT "matrix_uuid_for_models";
                  ALTER TABLE test_results.test_predictions DROP CONSTRAINT "matrix_uuid_for_testpred";
                  ALTER TABLE train_results.train_predictions DROP CONSTRAINT "matrix_uuid_for_trainpred"; ''')

    op.execute('ALTER TABLE model_metadata.models SET SCHEMA results;' +
              ' ALTER TABLE model_metadata.model_groups SET SCHEMA results;' +
              ' ALTER TABLE model_metadata.experiments SET SCHEMA results;' +
              ' ALTER TABLE model_metadata.list_predictions SET SCHEMA results;' +
              ' ALTER TABLE test_results.test_predictions RENAME TO predictions;' +
              ' ALTER TABLE test_results.predictions SET SCHEMA results;' +
              ' ALTER TABLE test_results.test_evaluations RENAME TO evaluations;' +
              ' ALTER TABLE test_results.evaluations SET SCHEMA results;' +
              ' ALTER TABLE test_results.individual_importances SET SCHEMA results;' +
              ' ALTER TABLE train_results.feature_importances SET SCHEMA results;')

    op.drop_index(op.f('ix_model_metadata_matrices_matrix_uuid'), table_name='matrices', schema='model_metadata')
    op.execute('ALTER TABLE results.models DROP COLUMN IF EXISTS model_size')
    op.execute('DROP TABLE model_metadata.matrices CASCADE')
    op.execute('DROP TABLE train_results.train_predictions CASCADE')
    op.execute('DROP TABLE train_results.train_evaluations CASCADE')

    op.execute('DROP SCHEMA IF EXISTS model_metadata')
    op.execute('DROP SCHEMA IF EXISTS train_results')
    op.execute('DROP SCHEMA IF EXISTS test_results')
    # ### end Alembic commands ###



"""Change importance to float

Revision ID: 72ac5cbdca05
Revises: 264245ddfce2
Create Date: 2017-09-01 14:31:09.302828

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '72ac5cbdca05'
down_revision = '264245ddfce2'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        table_name='individual_importances',
        column_name='importance_score',
        type_=sa.Float(),
        schema='results',
        postgresql_using='importance_score::double precision'
    )


def downgrade():
    op.alter_column(
        table_name='individual_importances',
        column_name='importance_score',
        type_=sa.Text(),
        schema='results'
    )



"""empty message

Revision ID: 0d44655e35fd
Revises: 8b3f167d0418
Create Date: 2017-08-31 14:14:54.187073

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0d44655e35fd'
down_revision = '8b3f167d0418'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'individual_importances',
        sa.Column('model_id', sa.Integer(), nullable=False),
        sa.Column('entity_id', sa.BigInteger(), nullable=False),
        sa.Column('as_of_date', sa.DateTime(), nullable=False),
        sa.Column('feature', sa.String(), nullable=False),
        sa.Column('method', sa.String(), nullable=False),
        sa.Column('importance_score', sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(['model_id'], ['results.models.model_id'], ),
        sa.PrimaryKeyConstraint(
            'model_id',
            'entity_id',
            'as_of_date',
            'feature',
            'method',
        ),
        schema='results'
    )
    op.create_table(
        'list_predictions',
        sa.Column('model_id', sa.Integer(), nullable=False),
        sa.Column('entity_id', sa.BigInteger(), nullable=False),
        sa.Column('as_of_date', sa.DateTime(), nullable=False),
        sa.Column('score', sa.Numeric(), nullable=True),
        sa.Column('rank_abs', sa.Integer(), nullable=True),
        sa.Column('rank_pct', sa.Float(), nullable=True),
        sa.Column('matrix_uuid', sa.Text(), nullable=True),
        sa.Column('test_label_window', sa.Interval(), nullable=True),
        sa.ForeignKeyConstraint(['model_id'], ['results.models.model_id'], ),
        sa.PrimaryKeyConstraint('model_id', 'entity_id', 'as_of_date'),
        schema='results'
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('list_predictions', schema='results')
    op.drop_table('individual_importances', schema='results')
    # ### end Alembic commands ###



import os
import tempfile
import unittest
import yaml
from collections import OrderedDict

import pandas as pd
from moto import mock_s3

from triage.component.catwalk.storage import (
    CSVMatrixStore,
    FSStore,
    HDFMatrixStore,
    S3Store,
    MatrixStorageEngine,
    ProjectStorage,
)


class SomeClass(object):
    def __init__(self, val):
        self.val = val


def test_S3Store():
    with mock_s3():
        import boto3
        client = boto3.client('s3')
        client.create_bucket(Bucket='test_bucket', ACL='public-read-write')
        store = S3Store(f"s3://test_bucket/a_path")
        assert not store.exists()
        store.write('val'.encode('utf-8'))
        assert store.exists()
        newVal = store.load()
        assert newVal.decode('utf-8') == 'val'
        store.delete()
        assert not store.exists()


def test_FSStore():
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpfile = os.path.join(tmpdir, 'tmpfile')
        store = FSStore(tmpfile)
        assert not store.exists()
        store.write('val'.encode('utf-8'))
        assert store.exists()
        newVal = store.load()
        assert newVal.decode('utf-8') == 'val'
        store.delete()
        assert not store.exists()


class MatrixStoreTest(unittest.TestCase):
    data_dict = OrderedDict([
        ('entity_id', [1, 2]),
        ('k_feature', [0.5, 0.4]),
        ('m_feature', [0.4, 0.5]),
        ('label', [0, 1])
    ])

    metadata = {
        'label_name': 'label',
        'indices': ['entity_id'],
    }

    def matrix_stores(self):
        df = pd.DataFrame.from_dict(self.data_dict).set_index(['entity_id'])

        with tempfile.TemporaryDirectory() as tmpdir:
            project_storage = ProjectStorage(tmpdir)
            tmpcsv = os.path.join(tmpdir, 'df.csv')
            tmpyaml = os.path.join(tmpdir, 'df.yaml')
            tmphdf = os.path.join(tmpdir, 'df.h5')
            with open(tmpyaml, 'w') as outfile:
                yaml.dump(self.metadata, outfile, default_flow_style=False)
                df.to_csv(tmpcsv)
                df.to_hdf(tmphdf, 'matrix')
                csv = CSVMatrixStore(project_storage, [], 'df')
                hdf = HDFMatrixStore(project_storage, [], 'df')
                assert csv.matrix.equals(hdf.matrix)
                yield from [csv, hdf]

    def test_MatrixStore_empty(self):
        for matrix_store in self.matrix_stores():
            assert not matrix_store.empty

    def test_MatrixStore_metadata(self):
        for matrix_store in self.matrix_stores():
            assert matrix_store.metadata == self.metadata

    def test_MatrixStore_head_of_matrix(self):
        for matrix_store in self.matrix_stores():
            assert matrix_store.head_of_matrix.to_dict() == {
                'k_feature': {1: 0.5},
                'm_feature': {1: 0.4},
                'label': {1: 0}
            }

    def test_MatrixStore_columns(self):
        for matrix_store in self.matrix_stores():
            assert matrix_store.columns() == ['k_feature', 'm_feature']

    def test_MatrixStore_resort_columns(self):
        for matrix_store in self.matrix_stores():
            result = matrix_store.\
                matrix_with_sorted_columns(
                    ['m_feature', 'k_feature']
                )\
                .values\
                .tolist()
            expected = [
                [0.4, 0.5],
                [0.5, 0.4]
            ]
            self.assertEqual(expected, result)

    def test_MatrixStore_already_sorted_columns(self):
        for matrix_store in self.matrix_stores():
            result = matrix_store.\
                matrix_with_sorted_columns(
                    ['k_feature', 'm_feature']
                )\
                .values\
                .tolist()
            expected = [
                [0.5, 0.4],
                [0.4, 0.5]
            ]
            self.assertEqual(expected, result)

    def test_MatrixStore_sorted_columns_subset(self):
        with self.assertRaises(ValueError):
            for matrix_store in self.matrix_stores():
                matrix_store.\
                    matrix_with_sorted_columns(['m_feature'])\
                    .values\
                    .tolist()

    def test_MatrixStore_sorted_columns_superset(self):
        with self.assertRaises(ValueError):
            for matrix_store in self.matrix_stores():
                matrix_store.\
                    matrix_with_sorted_columns(
                        ['k_feature', 'l_feature', 'm_feature']
                    )\
                    .values\
                    .tolist()

    def test_MatrixStore_sorted_columns_mismatch(self):
        with self.assertRaises(ValueError):
            for matrix_store in self.matrix_stores():
                matrix_store.\
                    matrix_with_sorted_columns(
                        ['k_feature', 'l_feature']
                    )\
                    .values\
                    .tolist()

    def test_MatrixStore_save(self):
        for matrix_store in self.matrix_stores():
            original_dict = matrix_store.matrix.to_dict()
            matrix_store.save()
            assert matrix_store._load().to_dict() == original_dict

    def test_as_of_dates_entity_index(self):
        data = {
            'entity_id': [1, 2],
            'feature_one': [0.5, 0.6],
            'feature_two': [0.5, 0.6],
        }
        with tempfile.TemporaryDirectory() as tmpdir:
            project_storage = ProjectStorage(tmpdir)
            matrix_store = CSVMatrixStore(project_storage, [], 'test')
            matrix_store.matrix = pd.DataFrame.from_dict(data)
            matrix_store.metadata = {'end_time': '2016-01-01', 'indices': ['entity_id']}

            self.assertEqual(matrix_store.as_of_dates, ['2016-01-01'])

    def test_as_of_dates_entity_date_index(self):
        data = {
            'entity_id': [1, 2, 1, 2],
            'feature_one': [0.5, 0.6, 0.5, 0.6],
            'feature_two': [0.5, 0.6, 0.5, 0.6],
            'as_of_date': ['2016-01-01', '2016-01-01', '2017-01-01', '2017-01-01']
        }
        with tempfile.TemporaryDirectory() as tmpdir:
            project_storage = ProjectStorage(tmpdir)
            matrix_store = CSVMatrixStore(project_storage, [], 'test')
            matrix_store.matrix = pd.DataFrame.from_dict(data).set_index(['entity_id', 'as_of_date'])
            matrix_store.metadata = {'indices': ['entity_id', 'as_of_date']}

        self.assertEqual(matrix_store.as_of_dates, ['2016-01-01', '2017-01-01'])

    def test_s3_save(self):
        with mock_s3():
            import boto3
            client = boto3.client('s3')
            client.create_bucket(Bucket='fake-matrix-bucket', ACL='public-read-write')
            example = next(self.matrix_stores())
            project_storage = ProjectStorage('s3://fake-matrix-bucket')

            tosave = CSVMatrixStore(project_storage, [], 'test')
            tosave.matrix = example.matrix
            tosave.metadata = example.metadata
            tosave.save()

            tocheck = CSVMatrixStore(project_storage, [], 'test')
            assert tocheck.metadata == example.metadata
            assert tocheck.matrix.to_dict() == example.matrix.to_dict()



import logging
import textwrap


DEFAULT_LABEL_NAME = 'outcome'


class LabelGeneratorNoOp(object):
    def generate_all_labels(self, labels_table, as_of_dates, label_timespans):
        logging.warning('No label configuration is available, so no labels will be created')

    def generate(self, start_date, label_timespan, labels_table):
        logging.warning('No label configuration is available, so no labels will be created')

    def clean_up(self, labels_table_name):
        pass


class LabelGenerator(object):
    def __init__(self, db_engine, query, label_name=None):
        self.db_engine = db_engine

        # query is expected to select a number of entity ids
        # and an outcome for each given an as-of-date
        self.query = query
        self.label_name = label_name or DEFAULT_LABEL_NAME

    def _create_labels_table(self, labels_table_name):
        self.db_engine.execute(
            'drop table if exists {}'.format(labels_table_name)
        )
        self.db_engine.execute('''
            create table {} (
            entity_id int,
            as_of_date date,
            label_timespan interval,
            label_name varchar(30),
            label_type varchar(30),
            label int
        )'''.format(labels_table_name))

    def generate_all_labels(
        self,
        labels_table,
        as_of_dates,
        label_timespans,
    ):
        self._create_labels_table(labels_table)
        logging.info('Creating labels for %s as of dates and %s label timespans',
                     len(as_of_dates),
                     len(label_timespans))
        for as_of_date in as_of_dates:
            for label_timespan in label_timespans:
                logging.info('Generating labels for as of date %s and '
                             'label timespan %s', as_of_date, label_timespan)
                self.generate(
                    start_date=as_of_date,
                    label_timespan=label_timespan,
                    labels_table=labels_table,
                )
        nrows = [
            row[0] for row in
            self.db_engine.execute(
                'select count(*) from {}'.format(labels_table))
        ][0]
        if nrows == 0:
            logging.warning(
                'Done creating labels, but no rows in labels table!')
        else:
            logging.info('Done creating labels table %s: rows: %s', labels_table, nrows)

    def generate(self, start_date, label_timespan, labels_table):
        """Generate labels table using a query

        Parameters
        ----------
        start_date: str
            as of date
        label_timespan: str
            postgresql readable time interval
        labels_table: str
            name of labels table
        """
        # we want to apply the as-of-date and label in the database driver,
        # so replace the user {as_of_date} with the SQL %(as_of_date)

        query_with_db_variables = self.query.format(
            as_of_date=start_date,
            label_timespan=label_timespan
        )

        full_insert_query = textwrap.dedent('''
            insert into {labels_table}
            select
                entities_and_outcomes.entity_id,
                '{as_of_date}' as as_of_date,
                '{label_timespan}'::interval as label_timespan,
                '{label_name}' as label_name,
                'binary' as label_type,
                entities_and_outcomes.outcome as label
            from ({user_query}) entities_and_outcomes
        ''').format(
            user_query=query_with_db_variables,
            labels_table=labels_table,
            as_of_date=start_date,
            label_timespan=label_timespan,
            label_name=self.label_name,
        )

        logging.debug("Running label creation query")
        logging.debug(full_insert_query)
        self.db_engine.execute(
            full_insert_query,
        )

    def clean_up(self, labels_table_name):
        self.db_engine.execute('drop table if exists {}'.format(labels_table_name))



# -*- coding: utf-8 -*-
"""test_collate

Unit tests for `collate` module.

"""
from triage.component.collate import Aggregate, Aggregation


def test_aggregate():
    agg = Aggregate("*", "count", {})
    assert list(map(str, agg.get_columns())) == ['count(*)']


def test_aggregate_when():
    agg = Aggregate("1", "count", {})
    assert list(
        map(str, agg.get_columns(when="date < '2012-01-01'"))
    ) == ["count(1) FILTER (WHERE date < '2012-01-01')"]


def test_ordered_aggregate():
    agg = Aggregate("", "mode", {}, "x")
    (expression,) = agg.get_columns()
    assert str(expression) == "mode() WITHIN GROUP (ORDER BY x)"
    assert expression.name == "x_mode"


def test_ordered_aggregate_when():
    agg = Aggregate("", "mode", {}, "x")
    assert list(
        map(str, agg.get_columns(when="date < '2012-01-01'"))
    ) == ["mode() WITHIN GROUP (ORDER BY x) FILTER (WHERE date < '2012-01-01')"]


def test_aggregate_tuple_quantity():
    agg = Aggregate(("x","y"), "corr", {})
    assert list(map(str, agg.get_columns())) == ["corr(x, y)"]


def test_aggregate_tuple_quantity_when():
    agg = Aggregate(("x","y"), "corr", {})
    assert list(
        map(str, agg.get_columns(when="date < '2012-01-01'"))
    ) == ["corr(x, y) FILTER (WHERE date < '2012-01-01')"]


def test_aggregate_imputation_lookup():
    agg = Aggregate("a", ["avg", "sum"], {
        "coltype": "aggregate",
        "avg": {"type": "mean"},
        "sum": {"type": "constant", "value": 3},
        "max": {"type": "zero"}
        })
    assert agg.column_imputation_lookup()['a_avg']['type'] == 'mean'
    assert agg.column_imputation_lookup()['a_avg']['coltype'] == 'aggregate'
    assert agg.column_imputation_lookup()['a_sum']['type'] == 'constant'
    assert agg.column_imputation_lookup()['a_sum']['value'] == 3
    assert agg.column_imputation_lookup()['a_sum']['coltype'] == 'aggregate'


def test_aggregate_imputation_lookup_all():
    agg = Aggregate("a", ["avg", "sum"], {
        "coltype": "aggregate",
        "all": {"type": "zero"},
        "sum": {"type": "constant", "value": 3},
        "max": {"type": "mean"}
        })
    assert agg.column_imputation_lookup()['a_avg']['type'] == 'zero'
    assert agg.column_imputation_lookup()['a_avg']['coltype'] == 'aggregate'
    assert agg.column_imputation_lookup()['a_sum']['type'] == 'constant'
    assert agg.column_imputation_lookup()['a_sum']['value'] == 3
    assert agg.column_imputation_lookup()['a_sum']['coltype'] == 'aggregate'


def test_aggregate_arithmetic():
    n = Aggregate("x", "sum", {})
    d = Aggregate("1", "count", {})
    m = Aggregate("y", "avg", {})

    (e,) = (n/d + m).get_columns(prefix="prefix_")
    assert str(e) == "((sum(x)*1.0 / count(1)) + avg(y))"
    assert e.name == "prefix_x_sum/1_count+y_avg"


def test_aggregate_format_kwargs():
    agg = Aggregate("'{collate_date}' - date", "min", {})
    assert list(
        map(str, agg.get_columns(format_kwargs={"collate_date":"2012-01-01"}))
    ) == ["min('2012-01-01' - date)"]


def test_aggregation_table_name_no_schema():
    # no schema
    assert Aggregation([], from_obj='source', groups=[], state_table='tbl')\
            .get_table_name() == '"source_aggregation"'
    assert Aggregation([], from_obj='source', groups=[], state_table='tbl')\
            .get_table_name(imputed=True) == '"source_aggregation_imputed"'

    # prefix
    assert Aggregation([], from_obj='source', prefix="mysource",
            groups=[], state_table='tbl')\
            .get_table_name() == '"mysource_aggregation"'
    assert Aggregation([], from_obj='source', prefix="mysource",
            groups=[], state_table='tbl')\
            .get_table_name(imputed=True) == '"mysource_aggregation_imputed"'

    # schema
    assert Aggregation([], from_obj='source', schema='schema',
            groups=[], state_table='tbl')\
            .get_table_name() == '"schema"."source_aggregation"'
    assert Aggregation([], from_obj='source', schema='schema',
            groups=[], state_table='tbl')\
            .get_table_name(imputed=True) == '"schema"."source_aggregation_imputed"'


def test_distinct():
    assert list(
        map(str, Aggregate("distinct x", "count", {}).get_columns())
    ) == ["count(distinct x)"]

    assert list(
        map(str, Aggregate("distinct x", "count", {}).get_columns(when="date < '2012-01-01'"))
    ) == ["count(distinct x) FILTER (WHERE date < '2012-01-01')"]

    assert list(
        map(str, Aggregate("distinct(x)", "count", {}).get_columns(when="date < '2012-01-01'"))
    ) == ["count(distinct (x)) FILTER (WHERE date < '2012-01-01')"]

    assert list(
        map(str, Aggregate("distinct(x,y)", "count", {}).get_columns(when="date < '2012-01-01'"))
    ) == ["count(distinct (x,y)) FILTER (WHERE date < '2012-01-01')"]



import itertools
import logging

from triage.util.conf import convert_str_to_relativedelta, dt_from_str

from . import utils

# Throughout the code here, we're going to follow an example
# based around the following config:
# {
#
#   feature_start_time: '1995-01-01',
#   feature_end_time: '2017-07-01',
#
#   label_start_time: '2012-01-01',
#   label_end_time: '2017-07-01',
#
#   model_update_frequency: '1year',
#
#   training_label_timespans: ['6month'],
#   test_label_timespans: ['6month'],
#
#   max_training_histories: ['2year'],
#   test_durations: ['3month'],
#
#   training_as_of_date_frequencies='1day',
#   test_as_of_date_frequencies='1month'
#
# }


class Timechop(object):

    def __init__(
        self,
        feature_start_time,
        feature_end_time,
        label_start_time,
        label_end_time,
        model_update_frequency,
        training_as_of_date_frequencies,
        max_training_histories,
        training_label_timespans,
        test_as_of_date_frequencies,
        test_durations,
        test_label_timespans
    ):
        self.feature_start_time = dt_from_str(feature_start_time)  # earliest time included in any feature
        self.feature_end_time = dt_from_str(feature_end_time)  # all data included in features are < this time
        if self.feature_start_time > self.feature_end_time:
            raise ValueError('Feature start time after feature end time.')

        self.label_start_time = dt_from_str(label_start_time)  # earliest time included in any label
        self.label_end_time = dt_from_str(label_end_time)  # all data in any label are < this time
        if self.label_start_time > self.label_end_time:
            raise ValueError('Label start time after label end time.')

        # how frequently to retrain models
        self.model_update_frequency = convert_str_to_relativedelta(model_update_frequency)

        # time between rows for same entity in train matrix
        self.training_as_of_date_frequencies = utils.convert_to_list(
            training_as_of_date_frequencies)

        # time between rows for same entity in test matrix
        self.test_as_of_date_frequencies = utils.convert_to_list(test_as_of_date_frequencies)

        # how much history for each entity to train on
        self.max_training_histories = utils.convert_to_list(max_training_histories)

        # how long into the future to make predictions for each entity
        self.test_durations = utils.convert_to_list(test_durations)

        # how much time is included in a label in the train matrix
        self.training_label_timespans = utils.convert_to_list(training_label_timespans)

        # how much time is included in a label in the test matrix
        self.test_label_timespans = utils.convert_to_list(test_label_timespans)

    def chop_time(self):
        """ Given the attributes of the object, define all train/test splits
        for all combinations of the temporal parameters.

        :return: a list of dictionaries defining train/test splits
        :rtype: list
        """
        matrix_set_definitions = []
        # in our example, we just have one value for each of these: 6month, 6month, and 3month
        for training_label_timespan, test_label_timespan, test_duration in itertools.product(
            self.training_label_timespans,
            self.test_label_timespans,
            self.test_durations
        ):
            # calculating the train-test split times starts from the end and walks backwards
            # e.g., train_test_split_times for our example with a 1 year model_update_frequency
            # will be every Oct. 1 from 2012 to 2016 (see comments in the method for details
            # on the calculation):
            # train_test_split_times = [2012-10-01, 2013-10-01, 2014-10-01, 2015-10-01, 2016-10-01]
            logging.info(
                'Calculating train/test split times for training prediction span {}, '
                'test prediction span {}, test span {}'.format(
                    training_label_timespan,
                    test_label_timespan,
                    test_duration
                )
            )
            train_test_split_times = self.calculate_train_test_split_times(
                training_label_timespan=convert_str_to_relativedelta(training_label_timespan),
                test_label_timespan=convert_str_to_relativedelta(test_label_timespan),
                test_duration=test_duration
            )
            logging.info('Train/test split times: {}'.format(train_test_split_times))

            # handle each training_as_of_date_frequency and max_training_history separately
            # to create matrices for each train_test_split_time.
            # in our example, we only have one value for each: 1day and 2year
            for training_as_of_date_frequency, max_training_history in itertools.product(
                self.training_as_of_date_frequencies,
                self.max_training_histories
            ):
                logging.info(
                    'Generating matrix definitions for training_as_of_date_frequency {}, '
                    'max_training_history {}'.format(
                        training_as_of_date_frequency,
                        max_training_history
                    )
                )
                for train_test_split_time in train_test_split_times:
                    logging.info(
                        'Generating matrix definitions for split {}'.format(
                            train_test_split_time
                        )
                    )
                    matrix_set_definitions.append(
                        self.generate_matrix_definitions(
                            train_test_split_time=train_test_split_time,
                            training_as_of_date_frequency=training_as_of_date_frequency,
                            max_training_history=max_training_history,
                            test_duration=test_duration,
                            training_label_timespan=training_label_timespan,
                            test_label_timespan=test_label_timespan
                        )
                    )
        return(matrix_set_definitions)

    def calculate_train_test_split_times(
        self,
        training_label_timespan,
        test_label_timespan,
        test_duration
    ):
        """ Calculate the split times between train and test matrices. All
        label spans in train matrices will end at this time, and this will be
        the first as of time in the respective test matrix.

        :param training_label_timespan: how much time is included in training
                                         labels
        :type training_label_timespan: dateutil.relativedelta.relativedelta
        :param test_label_timespan: how much time is included in test labels
        :type test_label_timespan: dateutil.relativedelta.relativedelta
        :param test_duration: for how long after the end of a training matrix are
                          test predictions made
        :type test_duration: str

        :return: all split times for the temporal parameters
        :rtype: list

        :raises: ValueError if there are no valid split times in the temporal
                 config
        """

        # we always want to be sure we're using the most recent data, so for the splits,
        # we start from the very end of time for which we have labels and walk backwards,
        # ensuring we leave enough of a buffer for the test_label_timespan to get a full
        # set of labels for our last testing as_of_date
        #
        # in our example, last_test_label_time = 2017-07-01 - 6month = 2017-01-01
        last_test_label_time = self.label_end_time - test_label_timespan

        # final label must be able to have feature data associated with it
        if last_test_label_time > self.feature_end_time:
            last_test_label_time = self.feature_end_time
            raise ValueError('Final test label date cannot be after end of feature time.')
        logging.info('Final label as of date: {}'.format(last_test_label_time))

        # all split times have to allow at least one training label before them
        # e.g., earliest_possible_split_time = max(1995-01-01, 2012-01-01) + 6month = 2012-01-01
        earliest_possible_split_time = (
            training_label_timespan +
            max(self.feature_start_time, self.label_start_time)
        )
        logging.info(
            'Earliest possible train/test split time: {}'.format(
                earliest_possible_split_time
            )
        )

        # last split is the first as of time in the final test matrix
        # that is, starting from the label_end_time, we've walked back by the test_label_timespan
        # (above) to allow a buffer for labels and now we walk back further by the test_duration to
        # ensure we have a full set of test data in the latest test matrix.
        #
        # e.g., last_split_time = 2017-01-01 - 3month = 2016-10-01
        test_delta = convert_str_to_relativedelta(test_duration)
        last_split_time = (
            last_test_label_time - test_delta
        )
        logging.info('Final split time: {}'.format(last_split_time))
        if last_split_time < earliest_possible_split_time:
            raise ValueError(
                'No valid train/test split times in temporal config.'
            )

        train_test_split_times = []
        train_test_split_time = last_split_time

        # finally, starting from our last_split_time, simply step backwards by the
        # model_update_frequency until we hit the earliest allowable time to
        # yield the set of train_test_split_times
        #
        # e.g., train_test_split_times for our example with a 1 year model_update_frequency
        # will be every Oct. 1 from 2012 to 2016:
        # train_test_split_times = [2012-10-01, 2013-10-01, 2014-10-01, 2015-10-01, 2016-10-01]
        while train_test_split_time >= earliest_possible_split_time:
            train_test_split_times.insert(0, train_test_split_time)
            train_test_split_time -= self.model_update_frequency

        return(train_test_split_times)

    # matrix_end_time is now matrix_end_time - label_window
    def calculate_as_of_times(
        self,
        as_of_start_limit,
        as_of_end_limit,
        data_frequency,
        forward=False
    ):
        """ Given a start and stop time, a frequncy, and a direction, calculate the
        as of times for a matrix.

        :param as_of_start_limit: the earliest possible as of time for a matrix
        :param as_of_end_limit: the last possible as of time for the matrix
        :param data_frequency: how much time between rows for a single entity
        :param forward: whether to generate times forward from the start time
                        (True) or backward from the end time (False)

        :return: list of as of times for the matrix
        :rtype: list
        """
        logging.info(
            'Calculating as_of_times from %s to %s using example frequency %s',
            as_of_start_limit,
            as_of_end_limit,
            data_frequency
        )

        as_of_times = []

        # in our example, this will apply to the test matrix with parameters
        #   as_of_start_limit = 2016-10-01, as_of_end_limit = 2017-01-01,
        #   data_frequency = 1month, forward=True
        # so, we'll start at 2016-10-01 and append this to the list of
        # as_of_times, then step forward one month at a time until we hit (but
        # do not include) 2017-01-01, yielding three values:
        #   [2016-10-01, 2016-11-01, 2016-12-01]
        if forward:
            as_of_time = as_of_start_limit
            # essentially a do-while loop for test matrices since
            # identical start and end times should include the first
            # date (e.g., ['2017-01-01', '2017-01-01') should give
            # preference to the inclusive side)
            as_of_times.append(as_of_time)
            as_of_time += data_frequency
            while as_of_time < as_of_end_limit:
                as_of_times.append(as_of_time)
                as_of_time += data_frequency

        # in our example, this will apply to the training matrix with parameters
        #   as_of_start_limit = 2014-04-01, as_of_end_limit = 2016-04-01,
        #   data_frequency = 1day, forward=False
        # so, we'll start from 2016-04-01 and step back by one day at a time
        # appending the results to the list of as_of_times until we hit 2014-04-01
        # (which will also be included)
        else:
            as_of_time = as_of_end_limit
            while as_of_time >= as_of_start_limit:
                as_of_times.insert(0, as_of_time)
                as_of_time -= data_frequency

        return(as_of_times)

    def generate_matrix_definitions(
        self,
        train_test_split_time,
        training_as_of_date_frequency,
        max_training_history,
        test_duration,
        training_label_timespan,
        test_label_timespan
    ):
        """ Given a split time and parameters for train and test matrices,
        generate as of times and metadata for the matrices in the split.

        :param train_test_split_time: the limit of the last label in the matrix
        :type train_test_split_time: datetime.datetime
        :param training_as_of_date_frequency: how much time between rows for an entity
                                        in a training matrix
        :type training_as_of_date_frequency: str
        :param max_training_history: how far back from split do train
                                    as_of_times go
        :type max_training_history: str
        :param test_duration: how far forward from split do test as_of_times go
        :type test_duration: str
        :param training_label_timespan: how much time covered by train labels
        :type training_label_timespan: str
        :param test_label_timespan: how much time is covered by test labels
        :type test_label_timespan: str

        :return: dictionary defining the train and test matrices for a split
        :rtype: dict
        """

        # continuing our example, let's consider the case when this is called for the last
        # train_test_split_time, so the parameters here are:
        #   train_test_split_time = 2016-10-01
        #   training_as_of_date_frequency = 1day
        #   max_training_history = 2year
        #   test_duration = 3month
        #   training_label_timespan = 6month
        #   test_label_timespan = 6month

        # for the example, the train matrix will contain as_of_dates for every day from
        # 2014-04-01 through 2016-04-01, including _both_ endpoints, providing a 6 month
        # buffer between the last as_of_time and the train-test split time for the last
        # set of labels (see comments in the method for details)
        train_matrix_definition = self.define_train_matrix(
            train_test_split_time=train_test_split_time,
            training_label_timespan=training_label_timespan,
            max_training_history=max_training_history,
            training_as_of_date_frequency=training_as_of_date_frequency
        )

        # for the example, the test matrix will contain three as_of_dates:
        #   [2016-10-01, 2016-11-01, 2016-12-01]
        # since we start at the train_test_split_time (2016-10-01) and walk forward by
        # the test_as_of_date_frequency (1 month) until we've exhausted the test_duration
        # (3 months), exclusive (see comments in the method for details)
        test_matrix_definitions = self.define_test_matrices(
            train_test_split_time=train_test_split_time,
            test_duration=test_duration,
            test_label_timespan=test_label_timespan
        )

        matrix_set_definition = {
            'feature_start_time': self.feature_start_time,
            'feature_end_time': self.feature_end_time,
            'label_start_time': self.label_start_time,
            'label_end_time': self.label_end_time,
            'train_matrix': train_matrix_definition,
            'test_matrices': test_matrix_definitions
        }
        logging.info(
            'Matrix definitions for train/test split {}: {}'.format(
                train_test_split_time,
                matrix_set_definition
            )
        )

        return(matrix_set_definition)

    def define_train_matrix(
        self,
        train_test_split_time,
        training_label_timespan,
        max_training_history,
        training_as_of_date_frequency
    ):
        """ Given a split time and the parameters of a training matrix, generate
        the as of times and metadata for a train matrix.

        :param train_test_split_time: the limit of the last label in the matrix
        :type train_test_split_time: datetime.datetime
        :param training_label_timespan: how much time is covered by the labels
        :type training_label_timespan: str
        :param max_training_history: how far back from split do as_of_times go
        :type max_training_history: str
        :param training_as_of_date_frequency: how much time between rows for an entity
        :type training_as_of_date_frequency: str

        :return: dictionary containing the temporal parameters and as of times
                 for a train matrix
        :rtype: dict
        """

        # for our example, this will be called with:
        #   train_test_split_time = 2016-10-01
        #   training_label_timespan = 6month
        #   max_training_history = 2year
        #   training_as_of_date_frequency = 1day

        # last as of time in the matrix is 1 label span before split to provide
        # enough of a buffer for the label data to avoid spilling into the test
        # matrix and causing a leakage problem.
        #
        # e.g., last_train_as_of_time = 2016-10-01 - 6month = 2016-04-01
        training_prediction_delta = convert_str_to_relativedelta(training_label_timespan)
        last_train_as_of_time = (
            train_test_split_time - training_prediction_delta
        )
        logging.info('last train as of time: {}'.format(last_train_as_of_time))

        # earliest time in matrix can't be farther back than the latest of the beginning
        # of label time or the beginning of feature time -- whichever is latest is the
        # limit if the amount of history we want to take would go further back.
        #
        # e.g., 2016-04-01 - 2year = 2014-04-01, which is later than both our
        # label_start_time (2012-01-01) and our feature_start_time (1995-01-01), so we
        # can use earliest_possible_train_as_of_time = 2014-04-01
        max_training_delta = convert_str_to_relativedelta(max_training_history)
        earliest_possible_train_as_of_time = (
            last_train_as_of_time - max_training_delta
        )
        experiment_as_of_time_limit = max(
            self.label_start_time,
            self.feature_start_time
        )
        if earliest_possible_train_as_of_time < experiment_as_of_time_limit:
            earliest_possible_train_as_of_time = experiment_as_of_time_limit
        logging.info(
            'earliest possible train as of time: {}'.format(
                earliest_possible_train_as_of_time
            )
        )

        # with the last as of time and the earliest possible time known,
        # calculate all the as of times for the matrix, stepping backwards
        # from the last as of time (to ensure that we use the latest possible
        # training data even if there's a gap and things don't line up
        # exactly) by the training_as_of_date_frequency
        #
        # for our example, this will give us a list of every day from 2014-04-01
        # through 2016-04-01, including _both_ endpoints
        train_as_of_times = self.calculate_as_of_times(
            as_of_start_limit=earliest_possible_train_as_of_time,
            as_of_end_limit=last_train_as_of_time,
            data_frequency=convert_str_to_relativedelta(training_as_of_date_frequency)
        )
        logging.info('train as of times: {}'.format(train_as_of_times))

        # create a dict of the matrix metadata
        matrix_definition = {
            'first_as_of_time': min(train_as_of_times),
            'last_as_of_time': max(train_as_of_times),
            'matrix_info_end_time': train_test_split_time,
            'as_of_times': train_as_of_times,
            'training_label_timespan': training_label_timespan,
            'training_as_of_date_frequency': training_as_of_date_frequency,
            'max_training_history': max_training_history
        }

        return(matrix_definition)

    def define_test_matrices(
        self,
        train_test_split_time,
        test_duration,
        test_label_timespan
    ):
        """ Given a train/test split time and a set of testing parameters,
        generate the metadata and as of times for the test matrices in a split.

        :param train_test_split_time: the limit of the last label in the matrix
        :type train_test_split_time: datetime.datetime
        :param test_duration: how far forward from split do test as_of_times go
        :type test_duration: str
        :param test_label_timespan: how much time is covered by test labels
        :type test_label_timespan: str

        :return: list of dictionaries defining the test matrices for a split
        :rtype: list
        """

        # for our example, this will be called with:
        #   train_test_split_time = 2016-10-01
        #   test_duration = 3month
        #   test_label_timespan = 6month

        # the as_of_time_limit is simply the split time plus the test_duration and we
        # can avoid checking here for any issues with the label_end_time or
        # feature_end_time since we've guaranteed that those limits would be
        # satisfied when we calculated the train_test_split_times initially
        #
        # for the example, as_of_time_limit = 2016-10-01 + 3month = 2017-01-01
        # (note as well that this will be treated as an _exclusive_ limit)
        logging.info(
            'Generating test matrix definitions for train/test split {}'.format(
                train_test_split_time
            )
        )
        test_definitions = []
        test_delta = convert_str_to_relativedelta(test_duration)
        as_of_time_limit = train_test_split_time + test_delta
        logging.info('All test as of times before {}'.format(as_of_time_limit))

        # calculate the as_of_times associated with each test data frequency
        # for our example, we just have one, 1month
        for test_as_of_date_frequency in self.test_as_of_date_frequencies:
            logging.info(
                'Generating test matrix definitions for test data frequency {}'.format(
                    test_as_of_date_frequency
                )
            )

            # for test as_of_times we step _forwards_ from the train_test_split_time
            # to ensure that we always have a prediction set made immediately after
            # training is done (so, the freshest possible predictions) even if the
            # frequency doesn't divide the test_duration evenly so there's a gap before
            # the as_of_time_limit
            #
            # for our example, this will give three as_of_dates:
            #   [2016-10-01, 2016-11-01, 2016-12-01]
            # since we start at the train_test_split_time (2016-10-01) and walk forward by
            # the test_as_of_date_frequency (1 month) until we've exhausted the test_duration
            # (3 months), exclusive (see comments in the method for details)
            test_as_of_times = self.calculate_as_of_times(
                as_of_start_limit=train_test_split_time,
                as_of_end_limit=as_of_time_limit,
                data_frequency=convert_str_to_relativedelta(test_as_of_date_frequency),
                forward=True
            )
            logging.info('Test as of times: {}'.format(test_as_of_times))
            test_definition = {
                'first_as_of_time': train_test_split_time,
                'last_as_of_time': max(test_as_of_times),
                'matrix_info_end_time':
                    max(test_as_of_times) +
                    convert_str_to_relativedelta(test_label_timespan),
                'as_of_times': test_as_of_times,
                'test_label_timespan': test_label_timespan,
                'test_as_of_date_frequency': test_as_of_date_frequency,
                'test_duration': test_duration
            }
            test_definitions.append(test_definition)
        return(test_definitions)



import os.path

from alembic.config import Config
from alembic import command
import yaml
from sqlalchemy.engine.url import URL


from .schema import (
    Base,
    Experiment,
    FeatureImportance,
    IndividualImportance,
    ListPrediction,
    Matrix,
    Model,
    ModelGroup,
    TestEvaluation,
    TrainEvaluation,
    TestPrediction,
    TrainPrediction
)


__all__ = (
    'Base',
    'Experiment',
    'FeatureImportance',
    'IndividualImportance',
    'ListPrediction',
    'Matrix',
    'Model',
    'ModelGroup',
    'TestEvaluation',
    'TrainEvaluation',
    'TestPrediction',
    'TrainPrediction',
    'mark_db_as_upgraded',
    'upgrade_db',
)


def _base_alembic_args(db_config_filename=None):
    path = os.path.abspath(__file__)
    dir_path = os.path.dirname(path)
    alembic_ini_path = os.path.join(dir_path, 'alembic.ini')
    base = ['-c', alembic_ini_path]
    if db_config_filename:
        base += ['-x', 'db_config_file={}'.format(db_config_filename)]

    return base


def upgrade_db(db_config_filename=None, db_engine=None):
    if db_config_filename:
        command.upgrade(alembic_config(db_config_filename=db_config_filename), 'head')
    elif db_engine:
        command.upgrade(alembic_config(dburl=db_engine.url), 'head')
    else:
        raise ValueError('Must pass either a db_config_filename or a db_engine')


REVISION_MAPPING = {
    'v1': '72ac5cbdca05',  # may add some logic to tag earlier ones if needed, could be 26 or 0d
    'v2': '72ac5cbdca05',
    'v3': '7d57d1cf3429',
    'v4': '89a8ce240bae',
    'v5': '2446a931de7a',
    'pre-v1': '8b3f167d0418'
}


def stamp_db(revision, db_config_filename):
    command.stamp(alembic_config(db_config_filename=db_config_filename), revision)


def alembic_config(db_config_filename=None, dburl=None):
    if db_config_filename:
        with open(db_config_filename) as f:
            config = yaml.load(f)
            dburl = URL(
                'postgres',
                host=config['host'],
                username=config['user'],
                database=config['db'],
                password=config['pass'],
                port=config['port'],
            )
    path = os.path.abspath(__file__)
    dir_path = os.path.dirname(path)
    alembic_ini_path = os.path.join(dir_path, 'alembic.ini')
    config = Config(alembic_ini_path)
    alembic_path = os.path.join(dir_path, 'alembic')
    config.set_main_option('script_location', alembic_path)
    config.attributes['url'] = dburl
    return config



from __future__ import with_statement

import os

import yaml
from alembic import context
from sqlalchemy import create_engine
from sqlalchemy import pool
from sqlalchemy.engine.url import URL

from triage.component.results_schema import Base


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.

url = None

if 'url' in config.attributes:
    url = config.attributes['url']

if not url:
    url = os.environ.get('DBURL', None)

if not url:
    db_config_file = (context.get_x_argument('db_config_file')
                      .get('db_config_file', None))
    if not db_config_file:
        raise ValueError('No database connection information found')

    with open(db_config_file) as fd:
        config = yaml.load(fd)
        url = URL(
            'postgres',
            host=config['host'],
            username=config['user'],
            database=config['db'],
            password=config['pass'],
            port=config['port'],
        )


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        version_table='results_schema_versions'
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    connectable = create_engine(
        url,
        poolclass=pool.NullPool
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table='results_schema_versions',
            include_schemas=True,
        )
        connection.execute('set search_path to "{}", public'.format('results'))

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# -*- coding: utf-8 -*-
"""test_imputations

Unit tests for `collate.imputations` module.

"""
import pytest

from triage.component.collate.imputations import *


def test_impute_flag():
    imp = BaseImputation(column='a', coltype='aggregate')
    assert imp.imputed_flag_sql() == 'CASE WHEN "a" IS NULL THEN 1 ELSE 0 END AS "a_imp" '


def test_impute_flag_categorical():
    imp = BaseImputation(column='a', coltype='categorical')
    assert imp.imputed_flag_sql() is None


def test_mean_imputation():
    imp = ImputeMean(column='a', coltype='aggregate')
    assert imp.to_sql() == 'COALESCE("a", AVG("a") OVER (), 0) AS "a" '

    imp = ImputeMean(column='a', coltype='aggregate', partitionby='date')
    assert imp.to_sql() == 'COALESCE("a", AVG("a") OVER (PARTITION BY date), 0) AS "a" '

    imp = ImputeMean(column='a', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a", AVG("a") OVER (), 0) AS "a" '

    imp = ImputeMean(column='a', coltype='aggregate', partitionby='date')
    assert imp.to_sql() == 'COALESCE("a", AVG("a") OVER (PARTITION BY date), 0) AS "a" '

    imp = ImputeMean(column='a__NULL_mean', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a__NULL_mean", 1) AS "a__NULL_mean" '


def test_constant_imputation():
    imp = ImputeConstant(column='a', coltype='aggregate', value=3.14)
    assert imp.to_sql() == 'COALESCE("a", 3.14) AS "a" '

    imp = ImputeConstant(column='a_myval_max', coltype='categorical', value='myval')
    assert imp.to_sql() == 'COALESCE("a_myval_max", 1) AS "a_myval_max" '

    imp = ImputeConstant(column='a_otherval_max', coltype='categorical', value='myval')
    assert imp.to_sql() == 'COALESCE("a_otherval_max", 0) AS "a_otherval_max" '

    imp = ImputeConstant(column='a__NULL_mean', coltype='categorical', value='myval')
    assert imp.to_sql() == 'COALESCE("a__NULL_mean", 1) AS "a__NULL_mean" '


def test_impute_zero():
    imp = ImputeZero(column='a', coltype='aggregate')
    assert imp.to_sql() == 'COALESCE("a", 0) AS "a" '

    imp = ImputeZero(column='a_myval_max', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a_myval_max", 0) AS "a_myval_max" '

    imp = ImputeZero(column='a_otherval_max', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a_otherval_max", 0) AS "a_otherval_max" '

    imp = ImputeZero(column='a__NULL_mean', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a__NULL_mean", 1) AS "a__NULL_mean" '


def test_impute_zero_noflag():
    imp = ImputeZeroNoFlag(column='a', coltype='aggregate')
    assert imp.to_sql() == 'COALESCE("a", 0) AS "a" '
    assert imp.imputed_flag_sql() is None
    assert imp.noflag

    imp = ImputeZeroNoFlag(column='a_myval_max', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a_myval_max", 0) AS "a_myval_max" '

    imp = ImputeZeroNoFlag(column='a_otherval_max', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a_otherval_max", 0) AS "a_otherval_max" '

    imp = ImputeZeroNoFlag(column='a__NULL_mean', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a__NULL_mean", 0) AS "a__NULL_mean" '


def test_impute_null_cat():
    imp = ImputeNullCategory(column='a_myval_max', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a_myval_max", 0) AS "a_myval_max" '

    imp = ImputeNullCategory(column='a_otherval_max', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a_otherval_max", 0) AS "a_otherval_max" '

    imp = ImputeNullCategory(column='a__NULL_mean', coltype='categorical')
    assert imp.to_sql() == 'COALESCE("a__NULL_mean", 1) AS "a__NULL_mean" '

    try:
        imp = ImputeNullCategory(column='a', coltype='aggregate')
        imp.to_sql()
        assert False
    except ValueError:
        assert True
    else:
        assert False


def test_impute_binary_mode():
    imp = ImputeBinaryMode(column='a', coltype='aggregate')
    assert imp.to_sql() == 'COALESCE("a", CASE WHEN AVG("a") OVER () > 0.5 THEN 1 ELSE 0 END, 0) AS "a" '

    imp = ImputeBinaryMode(column='a', coltype='aggregate', partitionby='date')
    assert imp.to_sql() == 'COALESCE("a", CASE WHEN AVG("a") OVER (PARTITION BY date) > 0.5 THEN 1 ELSE 0 END, 0) AS "a" '

    try:
        imp = ImputeBinaryMode(column='a', coltype='categorical')
        imp.to_sql()
        assert False
    except ValueError:
        assert True
    else:
        assert False


def test_impute_error():
    try:
        imp = ImputeError(column='a', coltype='aggregate')
        imp.to_sql()
        assert False
    except ValueError:
        assert True
    else:
        assert False




import importlib
import logging
from datetime import datetime
from textwrap import dedent

from sklearn.model_selection import ParameterGrid

from triage.component import architect
from triage.component import catwalk
from triage.component.timechop import Timechop

from triage.util.conf import convert_str_to_relativedelta
from triage.validation_primitives import (
    table_should_have_data,
    column_should_be_intlike,
    column_should_be_stringlike,
    column_should_be_timelike,
)


class Validator(object):

    def __init__(self, db_engine=None, strict=True):
        self.db_engine = db_engine
        self.strict = strict

    def run(self, *args, **kwargs):
        try:
            self._run(*args, **kwargs)
        except ValueError as e:
            if self.strict:
                raise ValueError(e)
            else:
                logging.warning('Validation error hit, not running in strict mode so continuing on: %s', str(e))

class TemporalValidator(Validator):

    def _run(self, temporal_config):
        def dt_from_str(dt_str):
            return datetime.strptime(dt_str, '%Y-%m-%d')
        splits = []
        try:
            chopper = Timechop(
                feature_start_time=dt_from_str(
                    temporal_config['feature_start_time']),
                feature_end_time=dt_from_str(
                    temporal_config['feature_end_time']),
                label_start_time=dt_from_str(
                    temporal_config['label_start_time']),
                label_end_time=dt_from_str(temporal_config['label_end_time']),
                model_update_frequency=temporal_config['model_update_frequency'],
                training_label_timespans=temporal_config['training_label_timespans'],
                test_label_timespans=temporal_config['test_label_timespans'],
                training_as_of_date_frequencies=temporal_config['training_as_of_date_frequencies'],
                test_as_of_date_frequencies=temporal_config['test_as_of_date_frequencies'],
                max_training_histories=temporal_config['max_training_histories'],
                test_durations=temporal_config['test_durations'],
            )
            splits = chopper.chop_time()
        except Exception as e:
            raise ValueError(dedent('''
            Section: temporal_config -
            Timechop could not produce temporal splits from config {}.
            Error: {}
            '''.format(temporal_config, e)))
        for split_num, split in enumerate(splits):
            if len(split['train_matrix']['as_of_times']) == 0:
                raise ValueError(dedent('''
                Section: temporal_config -
                Computed split {} has a train matrix with no as_of_times.
                '''.format(split)))

            # timechop computes the last time available to train data
            # and stores it in the matrix as 'matrix_info_end_time'
            # but to be more sure, let's double-check by comparing as_of_times
            # in the train and all associated test matrices
            train_max_data_time = max(split['train_matrix']['as_of_times']) +\
                convert_str_to_relativedelta(
                    split['train_matrix']['training_label_timespan'])

            for test_matrix in split['test_matrices']:
                if len(test_matrix['as_of_times']) == 0:
                    raise ValueError(dedent('''
                    Section: temporal_config -
                    Computed split {} has a test matrix with no as_of_times.
                    '''.format(split)))
                overlapping_times = [
                    as_of_time
                    for as_of_time in test_matrix['as_of_times']
                    if as_of_time < train_max_data_time
                ]
                if overlapping_times:
                    raise ValueError(dedent('''
                    Section: temporal_config -
                    Computed split index {} has a test matrix with as_of_times {}
                    < the maximum train as_of_time + train label timespan.
                    ({}). This is likely an error in timechop. See the
                    experiment's split_definitions[{}] for more information'''.format(
                        split_num,
                        overlapping_times,
                        train_max_data_time,
                        split_num
                    )))


class FeatureAggregationsValidator(Validator):
    def _validate_keys(self, aggregation_config):
        for key in [
            'from_obj',
            'intervals',
            'groups',
            'knowledge_date_column',
            'prefix'
        ]:
            if key not in aggregation_config:
                raise ValueError(dedent('''
                Section: feature_aggregations -
                '{} required as key: aggregation config: {}'''.format(key, aggregation_config)))

    def _validate_aggregates(self, aggregation_config):
        if 'aggregates' not in aggregation_config \
                and 'categoricals' not in aggregation_config \
                and 'array_categoricals' not in aggregation_config:
            raise ValueError(dedent('''
            Section: feature_aggregations -
            Need either aggregates, categoricals, or array_categoricals
            in {}'''.format(aggregation_config)))

    def _validate_categoricals(self, categoricals):
        conn = self.db_engine.connect()
        for categorical in categoricals:
            if 'choice_query' in categorical and 'choices' in categorical:
                raise ValueError(dedent('''
                Section: feature_aggregations -
                Both 'choice_query' and 'choices' specified for {}.
                Please only specify one.'''.format(categorical)))
            if not ('choice_query' in categorical or 'choices' in categorical):
                raise ValueError(dedent('''
                Section: feature_aggregations -
                Neither 'choice_query' and 'choices' specified for {}.
                Please specify one.'''.format(categorical)))
            if 'choice_query' in categorical:
                logging.info('Validating choice query')
                choice_query = categorical['choice_query']
                try:
                    conn.execute('explain {}'.format(choice_query))
                except Exception as e:
                    raise ValueError(dedent('''
                    Section: feature_aggregations -
                    choice query does not run.
                    choice query: "{}"
                    Full error: {}'''.format(choice_query, e)))

    def _validate_from_obj(self, from_obj):
        conn = self.db_engine.connect()
        logging.info('Validating from_obj')
        try:
            conn.execute('explain select * from {}'.format(from_obj))
        except Exception as e:
            raise ValueError(dedent('''
                Section: feature_aggregations -
                from_obj query does not run.
                from_obj: "{}"
                Full error: {}'''.format(from_obj, e)))

    def _validate_time_intervals(self, intervals):
        logging.info('Validating time intervals')
        for interval in intervals:
            if interval != 'all':
                # this function, used elsewhere to break up time intervals,
                # will throw an error if the interval can't be converted to a
                # relativedelta
                try:
                    convert_str_to_relativedelta(interval)
                except Exception as e:
                    raise ValueError(dedent('''
                    Section: feature_aggregations -
                    Time interval is invalid.
                    interval: "{}"
                    Full error: {}'''.format(interval, e)))

    def _validate_groups(self, groups):
        if 'entity_id' not in groups:
            raise ValueError(dedent('''
            Section: feature_aggregations -
            List of groups needs to include 'entity_id'.
            Passed list: {}'''.format(groups)))

    def _validate_imputation_rule(self, aggregate_type, impute_rule):
        """Validate the imputation rule for a given aggregation type."""
        # dictionary of imputation type : required parameters
        valid_imputations = {
            'all': {
                'mean': [],
                'constant': ['value'],
                'zero': [],
                'zero_noflag': [],
                'error': []
            },
            'aggregates': {
                'binary_mode': [],
            },
            'categoricals': {
                'null_category': []
            }
        }
        valid_imputations['array_categoricals'] = valid_imputations['categoricals']

        # the valid imputation rules for the specific aggregation type being checked
        valid_types = dict(
            valid_imputations['all'], **valid_imputations[aggregate_type])

        # no imputation rule was specified
        if 'type' not in impute_rule.keys():
            raise ValueError(dedent('''
            Section: feature_aggregations -
            Imputation type must be specified'''))

        # a rule was specified, but not valid for this type of aggregate
        if impute_rule['type'] not in valid_types.keys():
            raise ValueError(dedent('''
            Section: feature_aggregations -
            Invalid imputation type %s for %s''' % (impute_rule['type'], aggregate_type)))

        # check that all required parameters exist in the keys of the imputation rule
        required_params = valid_types[impute_rule['type']]
        for param in required_params:
            if param not in impute_rule.keys():
                raise ValueError(dedent('''
                Section: feature_aggregations -
                Missing param %s for %s''' % (param, impute_rule['type'])))

    def _validate_imputations(self, aggregation_config):
        """Validate the imputation rules in an aggregation config, looping
        through all three types of aggregates. Most of the work here is
        done by _validate_imputation_rule() to check the requirements of
        each imputation rule found
        """
        agg_types = ['aggregates', 'categoricals', 'array_categoricals']

        for agg_type in agg_types:
            # base_imp are the top-level rules, `such as aggregates_imputation`
            base_imp = aggregation_config.get(agg_type+'_imputation', {})

            # loop through the individual aggregates
            for agg in aggregation_config.get(agg_type, []):
                # combine any aggregate-level imputation rules with top-level ones
                imp_dict = dict(base_imp, **agg.get('imputation', {}))

                # imputation rules are metric-specific, so check each metric's rule
                for metric in agg['metrics']:
                    # metric rules may be defined by the metric name (e.g., 'max')
                    # or with the 'all' catch-all, with named metrics taking
                    # precedence. If we fall back to {}, the rule validator will
                    # error out on no metric found.
                    impute_rule = imp_dict.get(metric, imp_dict.get('all', {}))
                    self._validate_imputation_rule(agg_type, impute_rule)

    def _validate_aggregation(self, aggregation_config):
        logging.info('Validating aggregation config %s', aggregation_config)
        self._validate_keys(aggregation_config)
        self._validate_aggregates(aggregation_config)
        self._validate_categoricals(aggregation_config.get('categoricals', []))
        self._validate_from_obj(aggregation_config['from_obj'])
        self._validate_time_intervals(aggregation_config['intervals'])
        self._validate_groups(aggregation_config['groups'])
        self._validate_imputations(aggregation_config)

    def _run(self, feature_aggregation_config):
        """Validate a feature aggregation config applied to this object

        The validations range from basic type checks, key presence checks,
        as well as validating the sql in from objects.

        Args:
            feature_aggregation_config (list) all values, except for feature
                date, necessary to instantiate a collate.SpacetimeAggregation

        Raises: ValueError if any part of the config is found to be invalid
        """
        if not feature_aggregation_config:
            raise ValueError(dedent('''
            Section: feature_aggregations -
            Section not found. You must define feature aggregations.'''))
        for aggregation in feature_aggregation_config:
            self._validate_aggregation(aggregation)


class LabelConfigValidator(Validator):
    def _validate_query(self, query):
        if '{as_of_date}' not in query:
            raise ValueError(dedent('''
            Section: label_config -
            If 'query' is used as label_config,
            {as_of_date} must be present'''))
        if '{label_timespan}' not in query:
            raise ValueError(dedent('''
            Section: label_config -
            If 'query' is used as label_config,
            {label_timespan} must be present'''))
        bound_query = query\
            .replace('{as_of_date}', '2016-01-01')\
            .replace('{label_timespan}', '6month')
        conn = self.db_engine.connect()
        logging.info('Validating label query via EXPLAIN')
        try:
            conn.execute('explain {}'.format(bound_query))
        except Exception as e:
            raise ValueError(dedent('''
                Section: label_config -
                given query can not be run with a sample as_of_date and label_timespan.
                query: "{}"
                Full error: {}'''.format(query, e)))

    @staticmethod
    def _validate_include_missing_labels_in_train_as(missing_label_flag):
        if missing_label_flag not in {None, True, False}:
            raise ValueError(dedent('''
            Section: label_config -
            The value for 'include_missing_labels_in_train_as', {}, is invalid.
            The key must be either absent, or a boolean value True or False
            Triage only supports binary labels at this time.'''.format(missing_label_flag)))

    def _run(self, label_config):
        if not label_config:
            raise ValueError(dedent('''
            Section: label_config -
            Section not found. You must define a label config.'''))

        if 'query' not in label_config:
            raise ValueError(dedent('''
            Section: label_config -
            key 'query' not found. You must define a label query.'''))
        self._validate_query(label_config['query'])
        self._validate_include_missing_labels_in_train_as(
            label_config.get('include_missing_labels_in_train_as', None)
        )


class CohortConfigValidator(Validator):
    def _run(self, cohort_config):
        mutex_keys = set(['dense_states', 'entities_table', 'query'])
        available_keys = mutex_keys | set(['name'])
        used_keys = set(cohort_config.keys())
        bad_keys = used_keys - available_keys
        if bad_keys:
            raise ValueError(dedent('''
                Section: cohort_config -
                The following given keys: '{}'
                are invalid. Available keys are: '{}'
                '''.format(bad_keys, available_keys)))
        used_mutex_keys = mutex_keys & used_keys
        if len(used_mutex_keys) > 1:
            raise ValueError(dedent('''
                Section: cohort_config -
                Only one of the following keys can be sent: '{}'
                Found '{}'
                '''.format(mutex_keys, used_mutex_keys)))

        if 'dense_states' in cohort_config:
            state_config = cohort_config['dense_states']
            if 'table_name' not in state_config:
                raise ValueError(dedent('''
                Section: cohort_config -
                If 'dense_states' is used as cohort config,
                a table name must be present'''))
            dense_state_table = state_config['table_name']
            table_should_have_data(dense_state_table, self.db_engine)
            column_should_be_intlike(
                dense_state_table, 'entity_id', self.db_engine)
            column_should_be_stringlike(
                dense_state_table, 'state', self.db_engine)
            column_should_be_timelike(
                dense_state_table, 'start_time', self.db_engine)
            column_should_be_timelike(
                dense_state_table, 'end_time', self.db_engine)
            if 'state_filters' not in state_config or len(state_config['state_filters']) < 1:
                raise ValueError(dedent('''
                Section: cohort_config -
                If 'dense_states' is used as cohort config,
                at least one state filter must be present'''))
        elif 'entities_table' in cohort_config:
            entities_table = cohort_config['entities_table']
            table_should_have_data(entities_table, self.db_engine)
            column_should_be_intlike(
                entities_table, 'entity_id', self.db_engine)
        elif 'query' in cohort_config:
            query = cohort_config['query']
            if '{as_of_date}' not in query:
                raise ValueError(dedent('''
                Section: cohort_config -
                If 'query' is used as cohort_config,
                {as_of_date} must be present'''))
            dated_query = query.replace(
                '{as_of_date}', '2016-01-01')
            conn = self.db_engine.connect()
            logging.info('Validating cohort query')
            try:
                conn.execute(f'explain {dated_query}')
            except Exception as e:
                raise ValueError(dedent(f'''
                    Section: cohort_config -
                    given query can not be run with a sample as_of_date .
                    query: "{query}"
                    Full error: {e}'''))


class FeatureGroupDefinitionValidator(Validator):
    def _run(self, feature_group_definition, feature_aggregation_config):
        if not isinstance(feature_group_definition, dict):
            raise ValueError(dedent('''
            Section: feature_group_definition -
            feature_group_definition must be a dictionary'''))

        available_subsetters = architect.feature_group_creator.FeatureGroupCreator.subsetters
        for subsetter_name, value in feature_group_definition.items():
            if subsetter_name not in available_subsetters:
                raise ValueError(dedent('''
                Section: feature_group_definition -
                Unknown feature_group_definition key {} received.
                Available keys are {}'''.format(subsetter_name, available_subsetters)))
            if not hasattr(value, '__iter__') or isinstance(value, (str, bytes)):
                raise ValueError(dedent('''
                Section: feature_group_definition -
                feature_group_definition value for {}, {}
                should be a list'''.format(subsetter_name, value)))

        if 'prefix' in feature_group_definition:
            available_prefixes = {
                aggregation['prefix']
                for aggregation in feature_aggregation_config
            }
            bad_prefixes = set(
                feature_group_definition['prefix']) - available_prefixes
            if bad_prefixes:
                raise ValueError(dedent('''
                Section: feature_group_definition -
                The following given feature group prefixes: '{}'
                are invalid. Available prefixes from this experiment's feature
                aggregations are: '{}'
                '''.format(bad_prefixes, available_prefixes)))

        if 'tables' in feature_group_definition:
            available_tables = {
                aggregation['prefix'] + '_aggregation_imputed'
                for aggregation in feature_aggregation_config
            }
            bad_tables = set(
                feature_group_definition['tables']) - available_tables
            if bad_tables:
                raise ValueError(dedent('''
                Section: feature_group_definition -
                The following given feature group tables: '{}'
                are invalid. Available tables from this experiment's feature
                aggregations are: '{}'
                '''.format(bad_tables, available_tables)))


class FeatureGroupStrategyValidator(Validator):
    def _run(self, feature_group_strategies):
        if not isinstance(feature_group_strategies, list):
            raise ValueError(dedent('''
            Section: feature_group_strategies -
            feature_group_strategies section must be a list'''))
        available_strategies = {
            key for key in
            architect.feature_group_mixer.FeatureGroupMixer.strategy_lookup.keys()
        }
        bad_strategies = set(feature_group_strategies) - available_strategies
        if bad_strategies:
            raise ValueError(dedent('''
            Section: feature_group_strategies -
            The following given feature group strategies:
            '{}' are invalid. Available strategies are: '{}'
            '''.format(bad_strategies, available_strategies)))


class UserMetadataValidator(Validator):
    def _run(self, user_metadata):
        if not isinstance(user_metadata, dict):
            raise ValueError(dedent('''
            Section: user_metadata -
            user_metadata section must be a dict'''))


class ModelGroupKeysValidator(Validator):
    def _run(self, model_group_keys, user_metadata):
        if not isinstance(model_group_keys, list):
            raise ValueError(dedent('''
            Section: model_group_keys -
            model_group_keys section must be a list'''))
        classifier_keys = [
            'class_path',
            'parameters'
        ]
        # planner_keys are defined in architect.Planner._make_metadata
        planner_keys = [
            'feature_start_time',
            'end_time',
            'indices',
            'feature_names',
            'feature_groups',
            'label_name',
            'label_type',
            'label_timespan',
            'state',
            'cohort_name',
            'matrix_id',
            'matrix_type'
        ]
        # temporal_keys are defined in
        # timechop.Timechop.generate_matrix_definition
        temporal_keys = [
            'first_as_of_time',
            'last_as_of_time',
            'matrix_info_end_time',
            'as_of_times',
            'training_label_timespan',
            'training_as_of_date_frequency',
            'max_training_history'
        ]
        available_keys = [key for key in user_metadata.keys()] + \
            planner_keys +\
            temporal_keys +\
            classifier_keys
        for model_group_key in model_group_keys:
            if model_group_key not in available_keys:
                raise ValueError(dedent('''
                Section: model_group_keys -
                unknown entry '{}' received. Available keys are {}
                '''.format(model_group_key, available_keys)))


class GridConfigValidator(Validator):
    def _run(self, grid_config):
        if not grid_config:
            raise ValueError(dedent('''
            Section: grid_config -
            Section not found. You must define a grid_config.'''))
        for classpath, parameter_config in grid_config.items():
            try:
                module_name, class_name = classpath.rsplit(".", 1)
                module = importlib.import_module(module_name)
                cls = getattr(module, class_name)
                for parameters in ParameterGrid(parameter_config):
                    try:
                        cls(**parameters)
                    except Exception as e:
                        raise ValueError(dedent('''
                        Section: grid_config -
                        Unable to instantiate classifier {} with parameters {}, error thrown: {}
                        '''.format(classpath, parameters, e)))
            except Exception as e:
                raise ValueError(dedent('''
                Section: grid_config -
                Unable to import classifier {}, error thrown: {}
                '''.format(classpath, e)))


class ScoringConfigValidator(Validator):
    def _run(self, scoring_config):
        if 'testing_metric_groups' not in scoring_config:
            logging.warning('Section: scoring - No testing_metric_groups configured. ' +
                            'Your experiment may run, but you will not have any ' +
                            'evaluation metrics computed'
                            )
        if 'training_metric_groups' not in scoring_config:
            logging.warning('Section: scoring - No training_metric_groups configured. ' +
                            'If training set evaluation metrics are desired, they must be added'
                            )
        metric_lookup = catwalk.evaluation.ModelEvaluator.available_metrics
        available_metrics = set(metric_lookup.keys())
        for group in ('testing_metric_groups', 'training_metric_groups'):
            for metric_group in scoring_config.get(group, {}):
                given_metrics = set(metric_group['metrics'])
                bad_metrics = given_metrics - available_metrics
                if bad_metrics:
                    raise ValueError(dedent(
                        '''Section: scoring -
                        The following given metrics '{}' are unavailable. Available metrics are: '{}'
                        '''.format(bad_metrics, available_metrics)))
                for given_metric in given_metrics:
                    metric_function = metric_lookup[given_metric]
                    if not hasattr(metric_function, 'greater_is_better'):
                        raise ValueError(dedent(
                        '''Section: scoring -
                        The metric {} does not define the attribute
                        'greater_is_better'. This can only be fixed in the catwalk.metrics
                        module. If you still would like to use this metric, consider
                        submitting a pull request'''.format(given_metric)))


class ExperimentValidator(Validator):
    def run(self, experiment_config):
        TemporalValidator(strict=self.strict).run(experiment_config.get('temporal_config', {}))
        FeatureAggregationsValidator(self.db_engine, strict=self.strict)\
            .run(experiment_config.get('feature_aggregations', {}))
        LabelConfigValidator(self.db_engine, strict=self.strict).run(
            experiment_config.get('label_config', None))
        CohortConfigValidator(self.db_engine, strict=self.strict).run(
            experiment_config.get('cohort_config', {}))
        FeatureGroupDefinitionValidator(strict=self.strict).run(
            experiment_config.get('feature_group_definition', {}),
            experiment_config.get('feature_aggregations', {})
        )
        FeatureGroupStrategyValidator(strict=self.strict).run(
            experiment_config.get('feature_group_strategies', []),
        )
        UserMetadataValidator(strict=self.strict).run(
            experiment_config.get('user_metadata', {})
        )
        ModelGroupKeysValidator(strict=self.strict).run(
            experiment_config.get('model_group_keys', []),
            experiment_config.get('user_metadata', {})
        )
        GridConfigValidator(strict=self.strict).run(experiment_config.get('grid_config', {}))
        ScoringConfigValidator(strict=self.strict).run(experiment_config.get('scoring', {}))

        # show the success message in the console as well as the logger
        # as we don't really know how they have configured logging
        if self.strict:
            success_message = 'Experiment validation ran to completion with no errors'
        else:
            success_message = 'Experiment validation complete. All configuration problems have been displayed as warnings'
        logging.info(success_message)
        print(success_message)



import logging
from triage.component.architect.utils import str_in_sql


class FeatureDictionaryCreator(object):
    def __init__(self, features_schema_name, db_engine):
        self.features_schema_name = features_schema_name
        self.db_engine = db_engine

    def _tables_to_include(self, feature_table_names):
        return [
            feature_table for feature_table in feature_table_names
            if 'aggregation_imputed' in feature_table
        ]

    def feature_dictionary(self, feature_table_names, index_column_lookup):
        """ Create a dictionary of feature names, where keys are feature tables
        and values are lists of feature names.

        :return: feature_dictionary
        :rtype: dict
        """
        feature_dictionary = {}

        # iterate! store each table name + features names as key-value pair
        for feature_table_name in self._tables_to_include(feature_table_names):
            feature_names = [
                row[0] for row in
                self.db_engine.execute(
                    self._build_feature_names_query(
                        feature_table_name,
                        index_column_lookup[feature_table_name]
                    )
                )
            ]
            logging.info(
                'Feature names found for table %s: %s',
                feature_table_name,
                feature_names
            )
            feature_dictionary[feature_table_name] = feature_names
        return(feature_dictionary)

    def _build_feature_names_query(self, table_name, index_columns):
        """ For a given feature table, get the names of the feature columns.

        :param table_name: name of the feature table
        :type table_name: str

        :return: names of the feature columns in given table
        :rtype: list
        """
        # format the query that gets column names,
        # excluding indices from result
        feature_names_query = """
            SELECT column_name
            FROM information_schema.columns
            WHERE table_name = '{table}' AND
                  table_schema = '{schema}' AND
                  column_name NOT IN ({index_columns})
        """.format(
            table=table_name,
            schema=self.features_schema_name,
            index_columns=str_in_sql(index_columns)
        )
        logging.info(
            'Extracting all possible feature names for table %s with query %s',
            table_name,
            feature_names_query
        )

        return(feature_names_query)



import os
from datetime import datetime
from tempfile import TemporaryDirectory
import yaml


import testing.postgresql
from sqlalchemy import create_engine

from triage.component.results_schema import Base
from triage.component.timechop import Timechop
from triage.component.architect.features import (
    FeatureGenerator,
    FeatureDictionaryCreator,
    FeatureGroupCreator,
    FeatureGroupMixer,
)
from triage.component.architect.label_generators import LabelGenerator
from triage.component.architect.state_table_generators import StateTableGeneratorFromDense
from triage.component.architect.planner import Planner
from triage.component.architect.builders import MatrixBuilder
from triage.component.catwalk.storage import ProjectStorage

from tests.utils import sample_config


def populate_source_data(db_engine):
    cat_complaints = [
        (1, '2010-10-01', 5),
        (1, '2011-10-01', 4),
        (1, '2011-11-01', 4),
        (1, '2011-12-01', 4),
        (1, '2012-02-01', 5),
        (1, '2012-10-01', 4),
        (1, '2013-10-01', 5),
        (2, '2010-10-01', 5),
        (2, '2011-10-01', 5),
        (2, '2011-11-01', 4),
        (2, '2011-12-01', 4),
        (2, '2012-02-01', 6),
        (2, '2012-10-01', 5),
        (2, '2013-10-01', 6),
        (3, '2010-10-01', 5),
        (3, '2011-10-01', 3),
        (3, '2011-11-01', 4),
        (3, '2011-12-01', 4),
        (3, '2012-02-01', 4),
        (3, '2012-10-01', 3),
        (3, '2013-10-01', 4),
    ]

    dog_complaints = [
        (1, '2010-10-01', 5),
        (1, '2011-10-01', 4),
        (1, '2011-11-01', 4),
        (1, '2011-12-01', 4),
        (1, '2012-02-01', 5),
        (1, '2012-10-01', 4),
        (1, '2013-10-01', 5),
        (2, '2010-10-01', 5),
        (2, '2011-10-01', 5),
        (2, '2011-11-01', 4),
        (2, '2011-12-01', 4),
        (2, '2012-02-01', 6),
        (2, '2012-10-01', 5),
        (2, '2013-10-01', 6),
        (3, '2010-10-01', 5),
        (3, '2011-10-01', 3),
        (3, '2011-11-01', 4),
        (3, '2011-12-01', 4),
        (3, '2012-02-01', 4),
        (3, '2012-10-01', 3),
        (3, '2013-10-01', 4),
    ]

    events = [
        (1, 1, '2011-01-01'),
        (1, 1, '2011-06-01'),
        (1, 1, '2011-09-01'),
        (1, 1, '2012-01-01'),
        (1, 1, '2012-01-10'),
        (1, 1, '2012-06-01'),
        (1, 1, '2013-01-01'),
        (1, 0, '2014-01-01'),
        (1, 1, '2015-01-01'),
        (2, 1, '2011-01-01'),
        (2, 1, '2011-06-01'),
        (2, 1, '2011-09-01'),
        (2, 1, '2012-01-01'),
        (2, 1, '2013-01-01'),
        (2, 1, '2014-01-01'),
        (2, 1, '2015-01-01'),
        (3, 0, '2011-01-01'),
        (3, 0, '2011-06-01'),
        (3, 0, '2011-09-01'),
        (3, 0, '2012-01-01'),
        (3, 0, '2013-01-01'),
        (3, 1, '2014-01-01'),
        (3, 0, '2015-01-01'),
    ]

    states = [
        (1, 'state_one', '2012-01-01', '2016-01-01'),
        (1, 'state_two', '2013-01-01', '2016-01-01'),
        (2, 'state_one', '2012-01-01', '2016-01-01'),
        (2, 'state_two', '2013-01-01', '2016-01-01'),
        (3, 'state_one', '2012-01-01', '2016-01-01'),
        (3, 'state_two', '2013-01-01', '2016-01-01'),
    ]

    db_engine.execute('''create table cat_complaints (
        entity_id int,
        as_of_date date,
        cat_sightings int
        )''')

    for complaint in cat_complaints:
        db_engine.execute(
            "insert into cat_complaints values (%s, %s, %s)",
            complaint
        )

    db_engine.execute('''create table dog_complaints (
        entity_id int,
        as_of_date date,
        dog_sightings int
        )''')

    for complaint in dog_complaints:
        db_engine.execute(
            "insert into dog_complaints values (%s, %s, %s)",
            complaint
        )

    db_engine.execute('''create table events (
        entity_id int,
        outcome int,
        outcome_date date
    )''')

    for event in events:
        db_engine.execute(
            "insert into events values (%s, %s, %s)",
            event
        )

    db_engine.execute('''create table states (
        entity_id int,
        state text,
        start_time timestamp,
        end_time timestamp
    )''')

    for state in states:
        db_engine.execute(
            'insert into states values (%s, %s, %s, %s)',
            state
        )


def basic_integration_test(
    state_filters,
    feature_group_create_rules,
    feature_group_mix_rules,
    expected_matrix_multiplier,
    expected_group_lists
):
    with testing.postgresql.Postgresql() as postgresql:
        db_engine = create_engine(postgresql.url())
        Base.metadata.create_all(db_engine)
        populate_source_data(db_engine)

        with TemporaryDirectory() as temp_dir:
            chopper = Timechop(
                feature_start_time=datetime(2010, 1, 1),
                feature_end_time=datetime(2014, 1, 1),
                label_start_time=datetime(2011, 1, 1),
                label_end_time=datetime(2014, 1, 1),
                model_update_frequency='1year',
                training_label_timespans=['6months'],
                test_label_timespans=['6months'],
                training_as_of_date_frequencies='1day',
                test_as_of_date_frequencies='3months',
                max_training_histories=['1months'],
                test_durations=['1months'],
            )

            state_table_generator = StateTableGeneratorFromDense(
                db_engine=db_engine,
                experiment_hash='abcd',
                dense_state_table='states',
            )

            label_generator = LabelGenerator(
                db_engine=db_engine,
                query=sample_config()['label_config']['query']
            )

            feature_generator = FeatureGenerator(
                db_engine=db_engine,
                features_schema_name='features',
                replace=True,
            )

            feature_dictionary_creator = FeatureDictionaryCreator(
                db_engine=db_engine,
                features_schema_name='features'
            )

            feature_group_creator = FeatureGroupCreator(
                feature_group_create_rules)

            feature_group_mixer = FeatureGroupMixer(feature_group_mix_rules)
            project_storage = ProjectStorage(temp_dir)
            planner = Planner(
                feature_start_time=datetime(2010, 1, 1),
                label_names=['outcome'],
                label_types=['binary'],
                states=state_filters,
                user_metadata={},
            )

            builder = MatrixBuilder(
                engine=db_engine,
                db_config={
                    'features_schema_name': 'features',
                    'labels_schema_name': 'public',
                    'labels_table_name': 'labels',
                    'sparse_state_table_name': 'tmp_sparse_states_abcd',
                },
                matrix_storage_engine=project_storage.matrix_storage_engine(),
                replace=True
            )

            # chop time
            split_definitions = chopper.chop_time()
            num_split_matrices = sum(
                1 + len(split['test_matrices']) for split in split_definitions)

            # generate as_of_times for feature/label/state generation
            all_as_of_times = []
            for split in split_definitions:
                all_as_of_times.extend(split['train_matrix']['as_of_times'])
                for test_matrix in split['test_matrices']:
                    all_as_of_times.extend(test_matrix['as_of_times'])
            all_as_of_times = list(set(all_as_of_times))

            # generate sparse state table
            state_table_generator.generate_sparse_table(
                as_of_dates=all_as_of_times
            )

            # create labels table
            label_generator.generate_all_labels(
                labels_table='labels',
                as_of_dates=all_as_of_times,
                label_timespans=['6months']
            )

            # create feature table tasks
            # we would use FeatureGenerator#create_all_tables but want to use
            # the tasks dict directly to create a feature dict
            aggregations = feature_generator.aggregations(
                feature_aggregation_config=[{
                    'prefix': 'cat',
                    'from_obj': 'cat_complaints',
                    'knowledge_date_column': 'as_of_date',
                    'aggregates': [{
                        'quantity': 'cat_sightings',
                        'metrics': ['count', 'avg'],
                        'imputation': {
                            'all': {'type': 'mean'}
                        }
                    }],
                    'intervals': ['1y'],
                    'groups': ['entity_id']
                }, {
                    'prefix': 'dog',
                    'from_obj': 'dog_complaints',
                    'knowledge_date_column': 'as_of_date',
                    'aggregates_imputation': {
                        'count': {'type': 'constant', 'value': 7},
                        'sum': {'type': 'mean'},
                        'avg': {'type': 'zero'}
                    },
                    'aggregates': [{
                        'quantity': 'dog_sightings',
                        'metrics': ['count', 'avg'],
                    }],
                    'intervals': ['1y'],
                    'groups': ['entity_id']
                }],
                feature_dates=all_as_of_times,
                state_table=state_table_generator.sparse_table_name
            )
            feature_table_agg_tasks = feature_generator.generate_all_table_tasks(
                aggregations, task_type='aggregation')

            # create feature aggregation tables
            feature_generator.process_table_tasks(feature_table_agg_tasks)

            feature_table_imp_tasks = feature_generator.generate_all_table_tasks(
                aggregations, task_type='imputation')

            # create feature imputation tables
            feature_generator.process_table_tasks(feature_table_imp_tasks)

            # build feature dictionaries from feature tables and
            # subsetting config
            master_feature_dict = feature_dictionary_creator.feature_dictionary(
                feature_table_names=feature_table_imp_tasks.keys(),
                index_column_lookup=feature_generator.index_column_lookup(
                    aggregations)
            )

            feature_dicts = feature_group_mixer.generate(
                feature_group_creator.subsets(master_feature_dict)
            )

            # figure out what matrices need to be built
            _, matrix_build_tasks =\
                planner.generate_plans(
                    split_definitions,
                    feature_dicts
                )


            # go and build the matrices
            builder.build_all_matrices(matrix_build_tasks)

            # super basic assertion: did matrices we expect get created?
            matrices_records = list(db_engine.execute(
                    '''select matrix_uuid, num_observations, matrix_type
                    from model_metadata.matrices
                    '''
                ))
            matrix_directory = os.path.join(temp_dir, 'matrices')
            matrices = [path for path in os.listdir(
                matrix_directory) if '.csv' in path]
            metadatas = [path for path in os.listdir(
                matrix_directory) if '.yaml' in path]
            assert len(matrices) == num_split_matrices * \
                expected_matrix_multiplier
            assert len(metadatas) == num_split_matrices * \
                expected_matrix_multiplier
            assert len(matrices) == len(matrices_records)
            feature_group_name_lists = []
            for metadata_path in metadatas:
                with open(os.path.join(matrix_directory, metadata_path)) as f:
                    metadata = yaml.load(f)
                    feature_group_name_lists.append(metadata['feature_groups'])

            for matrix_uuid, num_observations, matrix_type in matrices_records:
                assert matrix_uuid in matrix_build_tasks #the hashes of the matrices
                assert type(num_observations) is int
                assert matrix_type == matrix_build_tasks[matrix_uuid]['matrix_type']
            def deep_unique_tuple(l): return set([tuple(i) for i in l])
            assert deep_unique_tuple(
                feature_group_name_lists) == deep_unique_tuple(expected_group_lists)


def test_integration_simple():
    matrices_records = basic_integration_test(
        state_filters=['state_one OR state_two'],
        feature_group_create_rules={'all': [True]},
        feature_group_mix_rules=['all'],
        # only looking at one state, and one feature group.
        # so we don't multiply timechop's output by anything
        expected_matrix_multiplier=1,
        expected_group_lists=[['all: True']],
    )



def test_integration_more_state_filtering():
    matrices_records = basic_integration_test(
        state_filters=['state_one OR state_two', 'state_one', 'state_two'],
        feature_group_create_rules={'all': [True]},
        feature_group_mix_rules=['all'],
        # 3 state filters, so the # of matrices should be each train/test split *3
        expected_matrix_multiplier=3,
        expected_group_lists=[['all: True']],
    )


def test_integration_feature_grouping():
    matrices_records = basic_integration_test(
        state_filters=['state_one OR state_two'],
        feature_group_create_rules={'prefix': ['cat', 'dog']},
        feature_group_mix_rules=['leave-one-out', 'all'],
        # 3 feature groups (cat/dog/cat+dog), so the # of matrices should be each train/test split *3
        expected_matrix_multiplier=3,
        expected_group_lists=[['prefix: cat'], [
            'prefix: cat', 'prefix: dog'], ['prefix: dog']],
    )



import datetime

from triage.component.architect import Planner
from triage.component.architect.feature_group_creator import FeatureGroup


def test_Planner():
    matrix_set_definitions = [
        {
            'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
            'modeling_start_time': datetime.datetime(2010, 1, 1, 0, 0),
            'modeling_end_time': datetime.datetime(2010, 1, 16, 0, 0),
            'train_matrix': {
                'first_as_of_time': datetime.datetime(2010, 1, 1, 0, 0),
                'matrix_info_end_time': datetime.datetime(2010, 1, 6, 0, 0),
                'as_of_times': [
                    datetime.datetime(2010, 1, 1, 0, 0),
                    datetime.datetime(2010, 1, 2, 0, 0),
                    datetime.datetime(2010, 1, 3, 0, 0),
                    datetime.datetime(2010, 1, 4, 0, 0),
                    datetime.datetime(2010, 1, 5, 0, 0)
                ]
            },
            'test_matrices': [{
                'first_as_of_time': datetime.datetime(2010, 1, 6, 0, 0),
                'matrix_info_end_time': datetime.datetime(2010, 1, 11, 0, 0),
                'as_of_times': [
                    datetime.datetime(2010, 1, 6, 0, 0),
                    datetime.datetime(2010, 1, 7, 0, 0),
                    datetime.datetime(2010, 1, 8, 0, 0),
                    datetime.datetime(2010, 1, 9, 0, 0),
                    datetime.datetime(2010, 1, 10, 0, 0)
                ]
            }]
        },
        {
            'feature_start_time': datetime.datetime(1990, 1, 1, 0, 0),
            'modeling_start_time': datetime.datetime(2010, 1, 1, 0, 0),
            'modeling_end_time': datetime.datetime(2010, 1, 16, 0, 0),
            'train_matrix': {
                'first_as_of_time': datetime.datetime(2010, 1, 6, 0, 0),
                'matrix_info_end_time': datetime.datetime(2010, 1, 11, 0, 0),
                'as_of_times': [
                    datetime.datetime(2010, 1, 6, 0, 0),
                    datetime.datetime(2010, 1, 7, 0, 0),
                    datetime.datetime(2010, 1, 8, 0, 0),
                    datetime.datetime(2010, 1, 9, 0, 0),
                    datetime.datetime(2010, 1, 10, 0, 0)
                ]
            },
            'test_matrices': [{
                'first_as_of_time': datetime.datetime(2010, 1, 11, 0, 0),
                'matrix_info_end_time': datetime.datetime(2010, 1, 16, 0, 0),
                'as_of_times': [
                    datetime.datetime(2010, 1, 11, 0, 0),
                    datetime.datetime(2010, 1, 12, 0, 0),
                    datetime.datetime(2010, 1, 13, 0, 0),
                    datetime.datetime(2010, 1, 14, 0, 0),
                    datetime.datetime(2010, 1, 15, 0, 0)
                ]
            }]
        }
    ]
    feature_dict_one = FeatureGroup(
        name='first_features',
        features_by_table={'features0': ['f1', 'f2'], 'features1': ['f1', 'f2']}
    )
    feature_dict_two = FeatureGroup(
        name='second_features',
        features_by_table={'features2': ['f3', 'f4'], 'features3': ['f5', 'f6']}
    )
    feature_dicts = [feature_dict_one, feature_dict_two]
    planner = Planner(
        feature_start_time=datetime.datetime(2010, 1, 1, 0, 0),
        label_names=['booking'],
        label_types=['binary'],
        cohort_name='prior_bookings',
        states=['state_one AND state_two'],
        user_metadata={},
    )

    updated_matrix_definitions, build_tasks = \
        planner.generate_plans(matrix_set_definitions, feature_dicts)
    # test that it added uuids: we don't much care what they are
    matrix_uuids = []
    for matrix_def in updated_matrix_definitions:
        assert isinstance(matrix_def['train_uuid'], str)
        matrix_uuids.append(matrix_def['train_uuid'])
        for test_uuid in matrix_def['test_uuids']:
            assert isinstance(test_uuid, str)
    assert len(set(matrix_uuids)) == 4

    # not going to assert anything on the keys (uuids), just get out the values
    build_tasks = build_tasks.values()
    assert len(build_tasks) == 8  # 2 splits * 2 matrices per split * 2 feature dicts

    assert sum(1 for task in build_tasks if task['matrix_type'] == 'train') == 4
    assert sum(1 for task in build_tasks if task['matrix_type'] == 'test') == 4
    assert sum(1 for task in build_tasks if task['feature_dictionary'] == feature_dict_one) == 4
    assert sum(1 for task in build_tasks if task['feature_dictionary'] == feature_dict_two) == 4
    assert sum(
        1 for task in build_tasks
        if task['matrix_metadata']['feature_groups'] == ['first_features']
    ) == 4
    assert sum(
        1 for task in build_tasks
        if task['matrix_metadata']['feature_groups'] == ['second_features']
    ) == 4
    assert sum(
        1 for task in build_tasks
        if task['matrix_metadata']['cohort_name'] == 'prior_bookings'
    ) == 8



from unittest.mock import patch

import numpy
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.audition.regrets import (
    SelectionRulePicker,
    SelectionRulePlotter,
)
from triage.component.audition.selection_rules import (
    best_current_value,
    best_average_value,
    BoundSelectionRule,
)

from .utils import create_sample_distance_table


def test_selection_rule_picker():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        distance_table, model_groups = create_sample_distance_table(engine)

        def pick_spiky(df, train_end_time):
            return [model_groups['spiky'].model_group_id]

        selection_rule_picker = SelectionRulePicker(
            distance_from_best_table=distance_table
        )

        results = selection_rule_picker.results_for_rule(
            bound_selection_rule=BoundSelectionRule(
                descriptive_name='spiky',
                function=pick_spiky,
                args={}
            ),
            model_group_ids=[mg.model_group_id for mg in model_groups.values()],
            train_end_times=['2014-01-01', '2015-01-01', '2016-01-01'],
            regret_metric='precision@',
            regret_parameter='100_abs',
        )
        assert [result['dist_from_best_case_next_time'] for result in results] == [0.19, 0.3, 0.12]
        assert [result['raw_value'] for result in results] == [0.45, 0.84, 0.45]


def test_selection_rule_picker_with_args():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        distance_table, model_groups = create_sample_distance_table(engine)

        def pick_highest_avg(df, train_end_time, metric, parameter):
            assert len(df['train_end_time'].unique()) == 2
            subsetted = df[(df['metric'] == metric) & (df['parameter'] == parameter)]
            mean = subsetted.groupby(['model_group_id'])['raw_value'].mean()
            return [mean.nlargest(1).index[0]]

        selection_rule_picker = SelectionRulePicker(
            distance_from_best_table=distance_table
        )
        regrets = [
            result['dist_from_best_case_next_time'] for result in
            selection_rule_picker.results_for_rule(
                bound_selection_rule=BoundSelectionRule(
                    descriptive_name='pick_highest_avg',
                    function=pick_highest_avg,
                    args={'metric': 'recall@', 'parameter': '100_abs'},
                ),
                model_group_ids=[mg.model_group_id for mg in model_groups.values()],
                train_end_times=['2015-01-01'],
                regret_metric='precision@',
                regret_parameter='100_abs',
            )
        ]
        # picking the highest avg recall will pick 'spiky' for this time
        assert regrets == [0.3]


def test_SelectionPlotter_create_plot_dataframe():
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        distance_table, model_groups = create_sample_distance_table(engine)
        plotter = SelectionRulePlotter(
            selection_rule_picker=SelectionRulePicker(distance_table)
        )
        plot_df = plotter.create_plot_dataframe(
            bound_selection_rules=[
                BoundSelectionRule(
                    descriptive_name='best_current_precision',
                    function=best_current_value,
                    args={'metric': 'precision@', 'parameter': '100_abs'}
                ),
                BoundSelectionRule(
                    descriptive_name='best_avg_precision',
                    function=best_average_value,
                    args={'metric': 'precision@', 'parameter': '100_abs'}
                ),
            ],
            model_group_ids=[mg.model_group_id for mg in model_groups.values()],
            train_end_times=['2014-01-01', '2015-01-01', '2016-01-01'],
            regret_metric='precision@',
            regret_parameter='100_abs',
        )
        # assert that we have the right # of columns and a row for each % diff value
        assert plot_df.shape == (100 * 2, 3)

        # both selection rules have a regret lower than 70
        for value in plot_df[plot_df['regret'] == 0.70]['pct_of_time'].values:
            assert numpy.isclose(value, 1.0)

        # best avg precision rule should be within 0.14 1/3 of the time
        for value in plot_df[
            (plot_df['regret'] == 0.14) &
            (plot_df['selection_rule'] == 'best_avg_precision')
        ]['pct_of_time'].values:
            assert numpy.isclose(value, 1.0/3)


def test_SelectionPlotter_plot():
    with patch('triage.component.audition.regrets.plot_cats') as plot_patch:
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            distance_table, model_groups = create_sample_distance_table(engine)
            plotter = SelectionRulePlotter(
                selection_rule_picker=SelectionRulePicker(distance_table)
            )
            plotter.plot_all_selection_rules(
                bound_selection_rules=[
                    BoundSelectionRule(
                        descriptive_name='best_current_precision',
                        function=best_current_value,
                        args={'metric': 'precision@', 'parameter': '100_abs'}
                    ),
                    BoundSelectionRule(
                        descriptive_name='best_avg_precision',
                        function=best_average_value,
                        args={'metric': 'precision@', 'parameter': '100_abs'}
                    ),
                ],
                model_group_ids=[mg.model_group_id for mg in model_groups.values()],
                train_end_times=['2014-01-01', '2015-01-01', '2016-01-01'],
                regret_metric='precision@',
                regret_parameter='100_abs',
            )
        assert plot_patch.called
        args, kwargs = plot_patch.call_args
        assert 'regret' in kwargs['frame']
        assert 'pct_of_time' in kwargs['frame']
        assert kwargs['x_col'] == 'regret'
        assert kwargs['y_col'] == 'pct_of_time'



from triage.component.catwalk.individual_importance import IndividualImportanceCalculator
from tests.utils import rig_engines, fake_trained_model, matrix_creator, matrix_metadata_creator, get_matrix_store

from unittest.mock import patch


def sample_individual_importance_strategy(
    db_engine,
    model_id,
    as_of_date,
    test_matrix_store,
    n_ranks
):
    return [{
        'entity_id': 1,
        'feature_value': 0.5,
        'feature_name': 'm_feature',
        'score': 0.5,
    }, {
        'entity_id': 1,
        'feature_value': 0.5,
        'feature_name': 'k_feature',
        'score': 0.5,
    }]


@patch.dict(
    'triage.component.catwalk.individual_importance.CALCULATE_STRATEGIES',
    {'sample': sample_individual_importance_strategy}
)
def test_calculate_and_save():
    with rig_engines() as (db_engine, project_storage):
        train_store = get_matrix_store(
            project_storage,
            matrix_creator(),
            matrix_metadata_creator(matrix_type='train'),
        )
        test_store = get_matrix_store(
            project_storage,
            matrix_creator(),
            matrix_metadata_creator(matrix_type='test'),
        )
        calculator = IndividualImportanceCalculator(db_engine, methods=['sample'], replace=False)
        # given a trained model
        # and a test matrix
        _, model_id = \
            fake_trained_model(
                db_engine,
                train_matrix_uuid=train_store.uuid
            )
        # i expect to be able to call calculate and save
        calculator.calculate_and_save_all_methods_and_dates(model_id, test_store)
        # and find individual importances in the results schema afterwards
        records = [
            row for row in
            db_engine.execute('''select entity_id, as_of_date
            from test_results.individual_importances
            join model_metadata.models using (model_id)''')
        ]
        assert len(records) > 0
        # and that when run again, has the same result
        calculator.calculate_and_save_all_methods_and_dates(model_id, test_store)
        new_records = [
            row for row in
            db_engine.execute('''select entity_id, as_of_date
            from test_results.individual_importances
            join model_metadata.models using (model_id)''')
        ]
        assert len(records) == len(new_records)
        assert records == new_records




class BaseImputation(object):
    """Base class for various imputation methods
    """
    def __init__(self, column, coltype, partitionby=None, null_cat_pattern=None, noflag=False):
        self.column = column
        self.coltype = coltype
        self.catcol = coltype in ['categorical', 'array_categorical']
        # categoricals have a null category, so don't need a flag
        self.noflag = True if self.catcol else noflag
        self.partitionby = "" if partitionby is None else "PARTITION BY %s" % partitionby
        # pattern for matching the null category column for a categorical variable
        # (assumes default of __NULL_ from collate.Compare):
        self.null_cat_pattern = '__NULL_' if null_cat_pattern is None else null_cat_pattern

    def _base_sql(self):
        return """COALESCE("{col}", {{imp}}) AS "{col}" """.format(col=self.column)

    def imputed_flag_sql(self):
        if not self.noflag:
            return """CASE WHEN "{col}" IS NULL THEN 1 ELSE 0 END AS "{col}_imp" """.format(
                col=self.column
            )
        else:
            # don't need to create a flag for categoricals (since this is handled with the
            # null category) or other imputations that suppress the flag (e.g., zero_noflag)
            return None


class ImputeMean(BaseImputation):
    """Class for mean imputation:

    For aggregate features, just take the average of the column (optionally within
    a partition, such as an as_of_date), falling back to 0 if the column is entirely
    null within the partition.

    For categorical features, we flag the NULL category column with a 1 and other
    columns with the mean, again falling back to 0 as necessary
    """
    def __init__(self, column, coltype, partitionby=None, null_cat_pattern=None, **kwargs):
        BaseImputation.__init__(
            self,
            column=column,
            coltype=coltype,
            partitionby=partitionby,
            null_cat_pattern=null_cat_pattern
        )

    def to_sql(self):
        sql = self._base_sql()

        if not self.catcol:
            # aggregate columm
            return sql.format(
                imp="""AVG("%s") OVER (%s), 0""" % (self.column, self.partitionby)
            )
        elif self.null_cat_pattern in self.column:
            # categorical NULL category
            return sql.format(imp=1)
        else:
            # categorical
            return sql.format(
                imp="""AVG("%s") OVER (%s), 0""" % (self.column, self.partitionby)
            )


class ImputeConstant(BaseImputation):
    """Class for constant value imputation:

    For aggregates, simply fill in the specified value.

    For categoricals, match the value to the column name and fill in the matching
    column with a 1 (as well as the NULL category column)
    """
    def __init__(self, column, coltype, value, null_cat_pattern=None, **kwargs):
        BaseImputation.__init__(
            self,
            column=column,
            coltype=coltype,
            partitionby=None,
            null_cat_pattern=null_cat_pattern
        )
        self.value = value

    def to_sql(self):
        sql = self._base_sql()

        if not self.catcol:
            # aggregate column
            return sql.format(imp=self.value)
        else:
            # categorical column
            return sql.format(
                imp=1 if self.value in self.column or self.null_cat_pattern in self.column else 0
            )


class ImputeZero(BaseImputation):
    """Class for zero filling imputation:

    Fill in the column with a 0, aside from a null category column for categorical
    variables, which is filled with a 1
    """
    def __init__(self, column, coltype, null_cat_pattern=None, **kwargs):
        BaseImputation.__init__(
            self,
            column=column,
            coltype=coltype,
            partitionby=None,
            null_cat_pattern=null_cat_pattern
        )

    def to_sql(self):
        sql = self._base_sql()
        return sql.format(
            imp=1 if self.catcol and self.null_cat_pattern in self.column else 0
        )


class ImputeZeroNoFlag(BaseImputation):
    """Class for zero filling with no imputation flag:

    Fill in missing values with 0 without generating an imputation flag. This option
    should be used only for cases where null values are explicitly known to be zero
    such as absence of an entity from an events table indicating that no such event
    has occurred.
    """
    def __init__(self, column, coltype, null_cat_pattern=None, **kwargs):
        BaseImputation.__init__(
            self,
            column=column,
            coltype=coltype,
            partitionby=None,
            null_cat_pattern=null_cat_pattern,
            noflag=True
        )

    def to_sql(self):
        sql = self._base_sql()
        return sql.format(
            imp=0
        )


class ImputeNullCategory(BaseImputation):
    """Class for just using the null category for categoricals:

    For a categorical feature, fill the null category with 1, all others with 0
    (essentially the same as ImputeZero for categoricals only)
    """
    def __init__(self, column, coltype, null_cat_pattern=None, **kwargs):
        BaseImputation.__init__(
            self,
            column=column,
            coltype=coltype,
            partitionby=None,
            null_cat_pattern=null_cat_pattern
        )
        if not self.catcol:
            raise ValueError('Can only use null category imputation for categorical features')

    def to_sql(self):
        sql = self._base_sql()
        return sql.format(
            imp=1 if self.null_cat_pattern in self.column else 0
        )


class ImputeBinaryMode(BaseImputation):
    """Class for mode imputation for binaries:

    For a binary variable, fill with a 1 if the mean (potentially in some window)
    is greater than 0.5, 0 otherwise. Note that this is not available for categoricals
    as it does not determine the modal category, just whether a binary is over 50%.
    """
    def __init__(self, column, coltype, partitionby=None, **kwargs):
        BaseImputation.__init__(
            self,
            column=column,
            coltype=coltype,
            partitionby=partitionby
        )
        if self.catcol:
            raise ValueError('Can only use binary mode imputation for non-categorical features')

    def to_sql(self):
        sql = self._base_sql()
        return sql.format(
            imp="""CASE WHEN AVG("%s") OVER (%s) > 0.5 THEN 1 ELSE 0 END, 0""" %
                (self.column, self.partitionby)
        )


class ImputeError(BaseImputation):
    """Class for raising an exception on null:

    If you expect your data to be clean, you can use this class to simply generate
    an error if any null values are found in the data rather than continuing with
    an imputation.
    """
    def __init__(self, column, coltype, **kwargs):
        BaseImputation.__init__(
            self,
            column=column,
            coltype=coltype
        )

    def to_sql(self):
        raise ValueError(
            "NULL values found in column with 'error' imputation type: %s" % self.column
            )



import pandas
import os
from .plotting import plot_cats


class SelectionRulePerformancePlotter(object):
    """Plot regrets over time

    Generates a simple line chart with audition.BoundSelectionRule objects as
    lines, time periods to be analyzed as the x-axis, and the regret of that
    selection rule as the y-axis.

    Args:
        selection_rule_picker (audition.SelectionRulePicker) An object that can
            simulate the effects of picking a selection rule
    """
    def __init__(self, selection_rule_picker, directory=None):
        self.selection_rule_picker = selection_rule_picker
        self.results_for_rule = None
        self.directory = directory

    def plot(
        self,
        bound_selection_rules,
        regret_metric,
        regret_parameter,
        model_group_ids,
        train_end_times,
        plot_type='regret'
    ):
        """Generate a selection rule performance plot for one metric

        Args:
            bound_selection_rules (list of .selection_rules.BoundSelectionRule)
               Selection rule/parameter combinations to plot
            regret_metric (string) -- model evaluation metric, such as 'precision@'
            regret_parameter (string) -- model evaluation metric parameter,
                    such as '300_abs'
            model_group_ids (list of integers) The model group ids to include
                in calculating the data
            train_end_times (list of timestamps) The timestamps to include in
                calculating the data
            plot_type (string) The plot type to show (either 'regret' or 'metric')
        """
        df = self.generate_plot_data(
            bound_selection_rules=bound_selection_rules,
            model_group_ids=model_group_ids,
            train_end_times=train_end_times,
            regret_metric=regret_metric,
            regret_parameter=regret_parameter,
        )
        if plot_type == 'regret':
            self.regret_plot_from_dataframe(
                metric=regret_metric,
                parameter=regret_parameter,
                df=df
            )
        elif plot_type == 'metric':
            self.raw_next_time_plot_from_dataframe(
                metric=regret_metric,
                parameter=regret_parameter,
                df=df
            )
        else:
            raise ValueError('Plot type must be either regret or metric')

    def generate_plot_data(
        self,
        bound_selection_rules,
        model_group_ids,
        train_end_times,
        regret_metric,
        regret_parameter
    ):
        """Create a dataframe suitable for plotting regrets over time

        Args:
            bound_selection_rules (list of .selection_rules.BoundSelectionRule)
               Selection rule/parameter combinations to plot
            model_group_ids (list of integers) The model group ids to include
                in calculating the data
            train_end_times (list of timestamps) The timestamps to include in
                calculating the data
            regret_metric (string) -- model evaluation metric, such as 'precision@'
            regret_parameter (string) -- model evaluation metric parameter,
                such as '300_abs'

        Returns: (pandas.DataFrame) A dataframe with columns 'regret',
            'train_end_time', and 'selection_rule'
        """
        accumulator = list()
        for selection_rule in bound_selection_rules:
            results = self.selection_rule_picker.results_for_rule(
                selection_rule,
                model_group_ids,
                train_end_times,
                regret_metric,
                regret_parameter,
            )
            for result in results:
                accumulator.append({
                    'train_end_time': result['train_end_time'],
                    'regret': result['dist_from_best_case_next_time'],
                    'selection_rule': selection_rule.descriptive_name,
                    'raw_value_next_time': result['raw_value_next_time'],
                    'model_group_id': result['model_group_id']
                })
        return pandas.DataFrame.from_records(accumulator)

    def regret_plot_from_dataframe(self, metric, parameter, df, **plt_format_args):
        """Generate a regret-over-time plot from a given dataframe

        Args:
            metric (string) -- model evaluation metric, such as 'precision@'
            parameter (string) -- model evaluation metric parameter,
                such as '300_abs'
            df (pandas.DataFrame) -- A dataframe preresenting regrets over time.
                 Should have columns 'regret', 'train_end_time', 'selection_rule'
        """
        cat_col = 'selection_rule'
        plt_title = 'Regret for {} {} over time'.format(metric, parameter)
        if self.directory:
            path_to_save = os.path.join(self.directory, f'regret_over_time_{metric}{parameter}.png')
        else:
            path_to_save = None
        plot_cats(
            frame=df,
            x_col='train_end_time',
            y_col='regret',
            cat_col=cat_col,
            grp_col='selection_rule',
            title=plt_title,
            x_label='train end time',
            y_label='regret in {}'.format(metric),
            x_lim=(df['train_end_time'].min(), df['train_end_time'].max()),
            path_to_save=path_to_save,
            **plt_format_args
        )

    def raw_next_time_plot_from_dataframe(self, metric, parameter, df, **plt_format_args):
        """Generate a 'raw-value-next-time' plot from a given dataframe

        Args:
            metric (string) -- model evaluation metric, such as 'precision@'
            parameter (string) -- model evaluation metric parameter,
                such as '300_abs'
            df (pandas.DataFrame) -- A dataframe that contains raw metric values
                over time given selection rules.
                Should have columns 'raw_value_next_time', 'train_end_time', 'selection_rule'
        """
        cat_col = 'selection_rule'
        plt_title = '{} {} next time'.format(metric, parameter)
        if self.directory:
            path_to_save = os.path.join(self.directory, f'{metric}{parameter}_next_time.png')
        else:
            path_to_save = None
        plot_cats(
            frame=df,
            x_col='train_end_time',
            y_col='raw_value_next_time',
            cat_col=cat_col,
            grp_col='selection_rule',
            title=plt_title,
            x_label='train end time',
            y_label='value of {} next time'.format(metric),
            x_lim=(df['train_end_time'].min(), df['train_end_time'].max()),
            path_to_save=path_to_save,
            **plt_format_args
        )



from triage.component.audition.distance_from_best import DistanceFromBestTable
from triage.component.catwalk.db import ensure_db

from tests.results_tests.factories import (
    ModelFactory,
    ModelGroupFactory,
    init_engine,
    session,
    MatrixFactory
)

def create_sample_distance_table(engine):
    ensure_db(engine)
    init_engine(engine)
    model_groups = {
        'stable': ModelGroupFactory(model_type='myStableClassifier'),
        'spiky': ModelGroupFactory(model_type='mySpikeClassifier'),
    }

    class StableModelFactory(ModelFactory):
        model_group_rel = model_groups['stable']

    class SpikyModelFactory(ModelFactory):
        model_group_rel = model_groups['spiky']

    models = {
        'stable_3y_ago': StableModelFactory(train_end_time='2014-01-01'),
        'stable_2y_ago': StableModelFactory(train_end_time='2015-01-01'),
        'stable_1y_ago': StableModelFactory(train_end_time='2016-01-01'),
        'spiky_3y_ago': SpikyModelFactory(train_end_time='2014-01-01'),
        'spiky_2y_ago': SpikyModelFactory(train_end_time='2015-01-01'),
        'spiky_1y_ago': SpikyModelFactory(train_end_time='2016-01-01'),
    }
    session.commit()
    distance_table = DistanceFromBestTable(
        db_engine=engine,
        models_table='models',
        distance_table='dist_table'
    )
    distance_table._create()
    stable_grp = model_groups['stable'].model_group_id
    spiky_grp = model_groups['spiky'].model_group_id
    stable_3y_id = models['stable_3y_ago'].model_id
    stable_3y_end = models['stable_3y_ago'].train_end_time
    stable_2y_id = models['stable_2y_ago'].model_id
    stable_2y_end = models['stable_2y_ago'].train_end_time
    stable_1y_id = models['stable_1y_ago'].model_id
    stable_1y_end = models['stable_1y_ago'].train_end_time
    spiky_3y_id = models['spiky_3y_ago'].model_id
    spiky_3y_end = models['spiky_3y_ago'].train_end_time
    spiky_2y_id = models['spiky_2y_ago'].model_id
    spiky_2y_end = models['spiky_2y_ago'].train_end_time
    spiky_1y_id = models['spiky_1y_ago'].model_id
    spiky_1y_end = models['spiky_1y_ago'].train_end_time
    distance_rows = [
        (stable_grp, stable_3y_id, stable_3y_end, 'precision@', '100_abs', 0.5, 0.6, 0.1, 0.5, 0.15),
        (stable_grp, stable_2y_id, stable_2y_end, 'precision@', '100_abs', 0.5, 0.84, 0.34, 0.5, 0.18),
        (stable_grp, stable_1y_id, stable_1y_end, 'precision@', '100_abs', 0.46, 0.67, 0.21, 0.5, 0.11),
        (spiky_grp, spiky_3y_id, spiky_3y_end, 'precision@', '100_abs', 0.45, 0.6, 0.15, 0.5, 0.19),
        (spiky_grp, spiky_2y_id, spiky_2y_end, 'precision@', '100_abs', 0.84, 0.84, 0.0, 0.5, 0.3),
        (spiky_grp, spiky_1y_id, spiky_1y_end, 'precision@', '100_abs', 0.45, 0.67, 0.22, 0.5, 0.12),
        (stable_grp, stable_3y_id, stable_3y_end, 'recall@', '100_abs', 0.4, 0.4, 0.0, 0.4, 0.0),
        (stable_grp, stable_2y_id, stable_2y_end, 'recall@', '100_abs', 0.5, 0.5, 0.0, 0.5, 0.0),
        (stable_grp, stable_1y_id, stable_1y_end, 'recall@', '100_abs', 0.6, 0.6, 0.0, 0.6, 0.0),
        (spiky_grp, spiky_3y_id, spiky_3y_end, 'recall@', '100_abs', 0.65, 0.65, 0.0, 0.65, 0.0),
        (spiky_grp, spiky_2y_id, spiky_2y_end, 'recall@', '100_abs', 0.55, 0.55, 0.0, 0.55, 0.0),
        (spiky_grp, spiky_1y_id, spiky_1y_end, 'recall@', '100_abs', 0.45, 0.45, 0.0, 0.45, 0.0),
    ]
    for dist_row in distance_rows:
        engine.execute(
            'insert into dist_table values (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)',
            dist_row
        )
    return distance_table, model_groups



import datetime
import unittest

from triage.util.conf import (
    convert_str_to_relativedelta,
    parse_delta_string,
)


class test_convert_str_to_relativedelta(unittest.TestCase):

    def test_valid_input(self):
        date = datetime.datetime(2016, 1, 1, 0, 0)
        for test in (
            {
                'interval': '1 year',
                'addition_result': datetime.datetime(2017, 1, 1, 0, 0),
                'subtraction_result': datetime.datetime(2015, 1, 1, 0, 0)
            },
            {
                'interval': '2 months',
                'addition_result': datetime.datetime(2016, 3, 1, 0, 0),
                'subtraction_result': datetime.datetime(2015, 11, 1, 0, 0)
            },
            {
                'interval': '3 days',
                'addition_result': datetime.datetime(2016, 1, 4, 0, 0),
                'subtraction_result': datetime.datetime(2015, 12, 29, 0, 0)
            },
            {
                'interval': '3 seconds',
                'addition_result': datetime.datetime(2016, 1, 1, 0, 0, 3),
                'subtraction_result': datetime.datetime(2015, 12, 31, 23, 59, 57)
            },
            {
                'interval': '2year',
                'addition_result': datetime.datetime(2018, 1, 1, 0, 0),
                'subtraction_result': datetime.datetime(2014, 1, 1, 0, 0)
            },
            {
                'interval': '4 weeks',
                'addition_result': datetime.datetime(2016, 1, 29, 0, 0),
                'subtraction_result': datetime.datetime(2015, 12, 4, 0, 0)
            },
            {
                'interval': '5 hours',
                'addition_result': datetime.datetime(2016, 1, 1, 5, 0),
                'subtraction_result': datetime.datetime(2015, 12, 31, 19, 0)
            },
            {
                'interval': '10minutes',
                'addition_result': datetime.datetime(2016, 1, 1, 0, 10),
                'subtraction_result': datetime.datetime(2015, 12, 31, 23, 50)
            },
            {
                'interval': '1microsecond',
                'addition_result': datetime.datetime(2016, 1, 1, 0, 0, 0, 1),
                'subtraction_result': datetime.datetime(2015, 12, 31, 23, 59, 59, 999999)
            },
            {
                'interval': '5m',
                'addition_result': datetime.datetime(2016, 1, 1, 0, 5),
                'subtraction_result': datetime.datetime(2015, 12, 31, 23, 55),
            },
        ):
            delta = convert_str_to_relativedelta(test['interval'])
            assert date + delta == test['addition_result']
            assert date - delta == test['subtraction_result']

    def test_bad_input(self):
        for bad_delta_string in ('4 tacos', '3'):
            with self.assertRaises(ValueError):
                convert_str_to_relativedelta(bad_delta_string)


class test_parse_delta_string(unittest.TestCase):
    def test_valid_input(self):
        delta_strings = ['1 year', '2 months', '3 days', '3s', '2y', '10m']
        expected_results = [
            ('year', 1),
            ('months', 2),
            ('days', 3),
            ('s', 3),
            ('y', 2),
            ('m', 10)
        ]
        for index, delta_string in enumerate(delta_strings):
            result = parse_delta_string(delta_string)
            assert result == expected_results[index]

    def test_invalid_input(self):
        delta_strings = ['one year', 'one_year', 'ay']
        for delta_string in delta_strings:
            with self.assertRaises(ValueError):
                parse_delta_string(delta_string)



"""
Upload files to s3
"""
import boto3
import botocore
import os
import glob
import logging


def upload_to_s3(access_key_id,
                 secret_access_key,
                 bucket,
                 folder,
                 ls_check_for_files=['matrix_pairs.txt', '.matrix_uuids'],
                 directory='.'):
    """
    Uploads files in a directory to a S3 bucket while checking for files
    metadatafiles and concatenating them.

    Parameters
    ----------
    access_key_id: str
        AWS access key
    secret_access_key: str
        AWS secret access key
    bucket: str
        Bucket name
    folder: str
        Folder to store data in S3 bucket
    ls_check_for_files: ls[str]
        Filenames to check for(default ['matching_pairs.txt','.pairs'])
    directory: str
        Directory to look for files in (default .)

    Returns
    -------
    logfile: file
        Logfile is written to /tmp/upload_s3.log
    """

    abs_path_dir = os.path.abspath(directory)

    logging.basicConfig(filename=abs_path_dir + '/upload_s3.log',
                        level=logging.DEBUG)

    s3 = boto3.client('s3',
                      aws_access_key_id=access_key_id,
                      aws_secret_access_key=secret_access_key)

    for filename in ls_check_for_files:
        try:
            s3.download_file(Bucket=bucket,
                             Key=folder + '/' + filename,
                             Filename=directory + '/' + "old_" + filename)

            if os.path.isfile("old_" + filename):
                abs_fname = directory + '/' + filename
                abs_oldfname = directory + '/' + "old_" + filename
                cmd = 'cat {} {} > tmp;mv tmp {};rm {}'.format(abs_fname,
                                                               abs_oldfname,
                                                               abs_fname,
                                                               abs_oldfname)
                logging.info(cmd)
                os.system(cmd)

        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                logging.info('File {} not in bucket'.format(filename))
        else:
            logging.info('Downloaded: {}'.format(filename))

    abs_path_check_files = [abs_path_dir + '/' +
                            fname for fname in ls_check_for_files]
    ls_files_to_upload = glob.glob(abs_path_dir + '/*') + abs_path_check_files
    for fname in ls_files_to_upload:
        logging.info('Uploading: {}'.format(fname))
        s3.upload_file(Bucket=bucket,
                       Key=folder + '/' + fname.split('/')[-1],
                       Filename=fname)



from sqlalchemy import Table
from sqlalchemy import create_engine
from sqlalchemy.types import VARCHAR
from testing.postgresql import Postgresql
from unittest import TestCase

from triage.component.architect import database_reflection as dbreflect


class TestDatabaseReflection(TestCase):

    def setUp(self):
        self.postgresql = Postgresql()
        self.engine = create_engine(self.postgresql.url())

    def tearDown(self):
        self.postgresql.stop()

    def test_split_table(self):
        assert dbreflect.split_table('staging.incidents') == ('staging', 'incidents')
        assert dbreflect.split_table('incidents') == (None, 'incidents')
        with self.assertRaises(ValueError):
            dbreflect.split_table('blah.staging.incidents')

    def test_table_object(self):
        assert isinstance(
            dbreflect.table_object('incidents', self.engine),
            Table
        )

    def test_reflected_table(self):
        self.engine.execute('create table incidents (col1 varchar)')
        assert dbreflect.reflected_table('incidents', self.engine).exists()

    def test_table_exists(self):
        self.engine.execute('create table incidents (col1 varchar)')
        assert dbreflect.table_exists('incidents', self.engine)
        assert not dbreflect.table_exists('compliments', self.engine)

    def test_table_has_data(self):
        self.engine.execute('create table incidents (col1 varchar)')
        self.engine.execute('create table compliments (col1 varchar)')
        self.engine.execute('insert into compliments values (\'good job\')')
        assert dbreflect.table_has_data('compliments', self.engine)
        assert not dbreflect.table_has_data('incidents', self.engine)

    def test_table_has_column(self):
        self.engine.execute('create table incidents (col1 varchar)')
        assert dbreflect.table_has_column('incidents', 'col1', self.engine)
        assert not dbreflect.table_has_column('incidents', 'col2', self.engine)

    def test_column_type(self):
        self.engine.execute('create table incidents (col1 varchar)')
        assert dbreflect.column_type('incidents', 'col1', self.engine) == VARCHAR
